from __future__ import annotations
from typing import Optional, TYPE_CHECKING
from aenum import Enum, NoAlias

# This is needed because for some reason pyright does not understand that Enum
# is a class (probably because Enum does weird things to the Enum class)
from typeguard import typechecked

if TYPE_CHECKING:
    from enum import Enum

from mpqp.core.instruction.measurement import Measure, ExpectationMeasure, BasisMeasure
from .connection.ibm_connection import get_IBMProvider, get_QiskitRuntimeService
from .connection.qlm_connection import get_QLMaaSConnection
from ..core.circuit import QCircuit
from .devices import AvailableDevice, IBMDevice, AWSDevice, ATOSDevice

from qat.comm.qlmaas.ttypes import JobStatus as QLM_JobStatus, QLMServiceException
from qiskit.providers import JobStatus as IBM_JobStatus
from braket.aws import AwsQuantumTask

from ..tools.errors import IBMRemoteExecutionError, QLMRemoteExecutionError


class JobStatus(Enum):
    """Possible states of a Job."""

    INIT = "initializing the job"
    """Initializing the job."""
    QUEUED = "the job is in the queue"
    """The job is in the queue."""
    RUNNING = "the job is currently running"
    """The job is currently running."""
    CANCELLED = "the job is cancelled"
    """The job is cancelled."""
    ERROR = "an error occurred with the job"
    """An error occurred with the job."""
    DONE = "the job is successfully done"
    """The job is successfully done."""


class JobType(Enum):
    """
    Possible types of Job to execute.

    Each type of job is restricted to some measures (and to some backends, but
    this is tackled by the backends themselves).
    """

    _settings_ = NoAlias
    STATE_VECTOR = {BasisMeasure, type(None)}
    """Retrieves the vector representing the quantum state, this type is *ideal*."""
    SAMPLE = {BasisMeasure}
    """Measures several times the quantum state in the basis, and retrieve the 
    counts. Contrarily to the ``STATE_VECTOR`` job type, this one is *realistic*."""
    OBSERVABLE = {ExpectationMeasure}
    """Computes the *expectation value* of an observable, using the state_vector 
    or the samples. This type is ideal too: it requires some trickery to 
    retrieve the expectation value in an optimal manner."""


@typechecked
class Job:
    """
    Representation of a job, an object regrouping all the information about
    the submission of a computation/measure of a quantum circuit on a
    specific hardware.

    A job has a type, and a status, and is attached to a specific device.
    Moreover, the job contains also the quantum circuit and the measure to be
    performed on the circuit.

    Args:
        job_type: Type of the job (sample, observable, ...).
        circuit: Circuit to execute.
        device: Device (simulator, quantum computer) on which we want to execute
            the job.
        measure: Object representing the measure to perform.

    Examples:
        >>> circuit = QCircuit(3)
        >>> job = Job(JobType.STATE_VECTOR, circuit, IBMDevice.AER_SIMULATOR)

        >>> circuit.add(BasisMeasure([0, 1], shots=1000))
        >>> job2 = Job(
        ...     JobType.STATE_VECTOR,
        ...     circuit,
        ...     IBMDevice.AER_SIMULATOR,
        ...     circuit.get_measurements()[0],
        ... )
    """

    # 6M-TODO: decide, when there are several measurements, if we define a
    #  multi-measure job, or if we need several jobs. For the moment, a Job can
    #  handle only one measurement

    def __init__(
        self,
        job_type: JobType,
        circuit: QCircuit,
        device: AvailableDevice,
        measure: Optional[Measure] = None,
    ):
        self._status = JobStatus.INIT

        self.job_type = job_type
        """See parameter description."""
        self.circuit = circuit
        """See parameter description."""
        self.device = device
        """See parameter description."""
        self.measure = measure
        """See parameter description."""

        self.id: Optional[str] = None
        """Contains the id of the remote job, used to retrieve the result from 
        the remote provider.  ``None`` if the job is local. If the job is not 
        local, it will be set later on."""

    @property
    def status(self):
        """Update and return the current job status. Mainly relevant for remote jobs."""
        if self._status not in [
            JobStatus.DONE,
            JobStatus.ERROR,
            JobStatus.CANCELLED,
        ]:
            # in the remote case, we need to check the current status of the job.
            # in the local case, it is updated automatically after each step
            if self.device.is_remote():
                assert isinstance(self.id, str)
                if isinstance(self.device, ATOSDevice):
                    self._status = get_qlm_job_status(self.id)

                elif isinstance(self.device, IBMDevice):
                    self._status = get_ibm_job_status(self.id)

                elif isinstance(self.device, AWSDevice):
                    self._status = get_aws_job_status(self.id)

                else:
                    raise NotImplementedError(
                        f"Cannot update job status for the device {self.device} yet"
                    )
        return self._status

    @status.setter
    def status(self, job_status: JobStatus):
        self._status = job_status


def get_qlm_job_status(job_id: str) -> JobStatus:
    """Retrieves the status of a QLM job from the id in parameter, and returns
    the corresponding JobStatus of this library.

    Args:
        job_id: Id of the job for which we want to retrieve the status.
    """
    try:
        qlm_status = get_QLMaaSConnection().get_status(job_id)
    except QLMServiceException as e:
        raise QLMRemoteExecutionError(
            f"Something went wrong when trying to get the state of job {job_id}."
            f" Please make sure the job_id is correct.\n Trace: " + str(e)
        )

    if qlm_status in [QLM_JobStatus.WAITING, QLM_JobStatus.STOPPED]:
        return JobStatus.QUEUED
    elif qlm_status == QLM_JobStatus.RUNNING:
        return JobStatus.RUNNING
    elif qlm_status == QLM_JobStatus.CANCELLED:
        return JobStatus.CANCELLED
    elif qlm_status in [
        QLM_JobStatus.FAILED,
        QLM_JobStatus.UNKNOWN_JOB,
        QLM_JobStatus.IN_BUCKET,
        QLM_JobStatus.DELETED,
    ]:
        return JobStatus.ERROR
    else:
        return JobStatus.DONE


def get_ibm_job_status(job_id: str) -> JobStatus:
    """Retrieves the status of an IBM job from the id in parameter, and returns
    the corresponding JobStatus of this library.

    Args:
        job_id: Id of the job for which we want to retrieve the status.
    """
    # test with QiskitRuntimeService
    if job_id in [e.job_id() for e in get_QiskitRuntimeService().jobs()]:
        ibm_job = get_QiskitRuntimeService().job(job_id)

    # if not, test with IBMProvider
    elif job_id in [e.job_id() for e in get_IBMProvider().jobs()]:
        ibm_job = get_IBMProvider().retrieve_job(job_id)

    else:
        raise IBMRemoteExecutionError(
            f"Could not find job with id {job_id} on IBM/QiskitRuntime provider"
        )
    status = ibm_job.status()
    if status == IBM_JobStatus.ERROR:
        return JobStatus.ERROR
    elif status == IBM_JobStatus.CANCELLED:
        return JobStatus.CANCELLED
    elif status in [IBM_JobStatus.QUEUED, IBM_JobStatus.VALIDATING]:
        return JobStatus.QUEUED
    elif status == IBM_JobStatus.INITIALIZING:
        return JobStatus.INIT
    elif status == IBM_JobStatus.RUNNING:
        return JobStatus.RUNNING
    else:
        return JobStatus.DONE


def get_aws_job_status(job_id: str) -> JobStatus:
    """Retrieves the status of a AWS Braket from the id in parameter, and
    returns the corresponding JobStatus of this library.

    Args:
        job_id: Id of the job for which we want to retrieve the status.
    """
    task = AwsQuantumTask(job_id)
    state = task.state()
    if state == "FAILED":
        return JobStatus.ERROR
    elif state == "CANCELLED":
        return JobStatus.CANCELLED
    elif state == "CREATED":
        return JobStatus.INIT
    elif state == "QUEUED":
        return JobStatus.QUEUED
    elif state == "RUNNING":
        return JobStatus.RUNNING
    else:
        return JobStatus.DONE



"""Example 3: Using native gates"""

from mpqp.gates import *
from mpqp import QCircuit
from mpqp.execution import run
from mpqp.execution.devices import IBMDevice, AWSDevice, ATOSDevice

# Declaration of the circuit with the right size
circuit = QCircuit(3, label="Test native gates")
# Constructing the circuit by adding gates and measurements
circuit.add([H(0), X(1), Y(2), Z(0), S(1), T(0)])
circuit.add([Rx(1.2324, 2), Ry(-2.43, 0), Rz(1.04, 1), Rk(-1, 1), P(-323, 2)])
circuit.add(U(1.2, 2.3, 3.4, 2))
circuit.add(SWAP(2, 0))
circuit.add([CNOT(0, 1), CRk(4, 2, 1), CZ(1, 2)])
circuit.add(TOF([0, 1], 2))

# no measure, we want the state vector

print(circuit)
print(circuit.to_qasm2())
print(circuit.to_qasm3())

result = run(circuit, [ATOSDevice.MYQLM_PYLINALG,
                       IBMDevice.AER_SIMULATOR_STATEVECTOR,
                       AWSDevice.BRAKET_LOCAL_SIMULATOR])
print(result)



"""Information about the state can be retrieved using the expectation value of
this state measured by an observable. This is done using the :class:`Observable`
class to define your observable, and a :class:`ExpectationMeasure` to perform
the measure."""

from __future__ import annotations

import copy
from numbers import Complex
from warnings import warn
from typing import Optional

import numpy as np
import numpy.typing as npt
from qiskit.circuit import Parameter
from sympy import Expr
from typeguard import typechecked

from mpqp.core.instruction.gates.native_gates import SWAP
from mpqp.core.instruction.measurement.measure import Measure
from mpqp.core.languages import Language
from mpqp.tools.errors import NumberQubitsError
from mpqp.tools.maths import is_hermitian
from mpqp.tools.generics import one_lined_repr


@typechecked
class Observable:
    """Class defining an observable, used for evaluating expectation values.

    An observable can be defined by using a hermitian matrix, or using a combination of operators in a specific
    basis (Kraus, Pauli, ...).

    For the moment, on can only define the observable using a matrix.

    Example:
        >>> matrix = np.array([[1, 0], [0, -1]])
        >>> obs = Observable(matrix)

    Args:
        matrix: Hermitian matrix representing the observable.
    """

    def __init__(self, matrix: npt.NDArray[np.complex64] | list[list[Complex]]):
        self.nb_qubits = int(np.log2(len(matrix)))
        """Number of qubits of this observable."""
        self.matrix = np.array(matrix)
        """See parameter description."""

        basis_states = 2**self.nb_qubits
        if self.matrix.shape != (basis_states, basis_states):
            raise ValueError(
                f"The size of the matrix {self.matrix.shape} doesn't neatly fit on a"
                " quantum register. It should be a square matrix of size a power"
                " of two."
            )

        if not is_hermitian(self.matrix):
            raise ValueError(
                "The matrix in parameter is not hermitian (cannot define an observable)"
            )

    def __repr__(self) -> str:
        return f"Observable({one_lined_repr(self.matrix)})"

    def __mult__(self, other: Expr | Complex) -> Observable:
        """3M-TODO"""
        ...

    def subs(
        self, values: dict[Expr | str, Complex], remove_symbolic: bool = False
    ) -> Observable:
        """3M-TODO"""
        ...


@typechecked
class ExpectationMeasure(Measure):
    """This measure evaluates the expectation value of the output of the circuit
    measured by the observable given as input.

    If the ``targets`` are not sorted and contiguous, some additional swaps will
    be needed. This will affect the performance of your circuit if run on noisy
    hardware. The swaps added can be checked out in the ``pre_measure``
    attribute of the :class:`ExpectationMeasure`.

    Example:
        >>> obs = Observable(np.diag([0.7, -1, 1, 1]))
        >>> c = QCircuit([H(0), CNOT(0,1), ExpectationMeasure([0,1], observable=obs, shots=10000)])
        >>> run(c, ATOSDevice.MYQLM_PYLINALG).expectation_value
        0.85918

    Args:
        targets: List of indices referring to the qubits on which the measure
            will be applied.
        observable: Observable used for the measure.
        shots: Number of shots to be performed.
        label: Label used to identify the measure.

    Warns:
        UserWarning: If the ``targets`` are not sorted and contiguous, some
            additional swaps will be needed. This will change the performance of
            your circuit is run on noisy hardware.

    Note:
        In a future version, we would will also allow you to provide a
        ``PauliDecomposition``, a decomposition of the observable in the
        generalized Pauli basis.
    """

    def __init__(
        self,
        targets: list[int],
        observable: Observable,
        shots: int = 0,
        label: Optional[str] = None,
    ):
        from mpqp.core.circuit import QCircuit

        super().__init__(targets, shots, label)
        self.observable = observable
        """See parameter description."""
        if self.nb_qubits != observable.nb_qubits:
            raise NumberQubitsError(
                f"{self.nb_qubits}, the number of target qubit(s) doesn't match"
                f" {observable.nb_qubits}, the size of the observable"
            )

        self.pre_measure = QCircuit(max(targets) + 1)
        """Circuit added before the expectation measurement to correctly swap
        target qubits when their are note ordered or contiguous."""
        targets_is_ordered = all(
            [targets[i] > targets[i - 1] for i in range(1, len(targets))]
        )
        tweaked_tgt = copy.copy(targets)
        if (
            max(tweaked_tgt) - min(tweaked_tgt) + 1 != len(tweaked_tgt)
            or not targets_is_ordered
        ):
            warn(
                "Non contiguous or non sorted observable target will introduce "
                "additional CNOTs."
            )

            for t_index, target in enumerate(tweaked_tgt):  # sort the targets
                min_index = tweaked_tgt.index(min(tweaked_tgt[t_index:]))
                if t_index != min_index:
                    self.pre_measure.add(SWAP(target, tweaked_tgt[min_index]))
                    tweaked_tgt[t_index], tweaked_tgt[min_index] = (
                        tweaked_tgt[min_index],
                        target,
                    )
            for t_index, target in enumerate(tweaked_tgt):  # compact the targets
                if t_index == 0:
                    continue
                if target != tweaked_tgt[t_index - 1] + 1:
                    self.pre_measure.add(SWAP(target, tweaked_tgt[t_index - 1] + 1))
                    tweaked_tgt[t_index] = tweaked_tgt[t_index - 1] + 1
        self.rearranged_targets = tweaked_tgt
        """Adjusted list of target qubits when they are not initially sorted and
        contiguous."""

    def __repr__(self) -> str:
        return (
            f"ExpectationMeasure({self.targets}, {self.observable}, shots={self.shots})"
        )

    def to_other_language(
        self,
        language: Language = Language.QISKIT,
        qiskit_parameters: Optional[set[Parameter]] = None,
    ) -> None:
        if qiskit_parameters is None:
            qiskit_parameters = set()
        if language == Language.QISKIT:
            raise NotImplementedError(
                "Qiskit does not implement these kind of measures"
            )
        else:
            raise NotImplementedError(
                "Only Qiskit supported for language export for now"
            )



"""Some gate (such as CNOT for instance) do not need any parameters, but in
order to have a universal set of gates, one needs at least one parametrized
gate. This module defines the abstract class needed to define these gates as
well as a way to handle symbolic variables.

More on the topic of symbolic variable can be found in the `VQA <VQA.html>`_
page"""

from __future__ import annotations

from abc import ABC
from copy import deepcopy
from typing import Optional
from numbers import Complex

from sympy import Expr, symbols  # pyright: ignore [reportUnusedImport]
from typeguard import typechecked

from mpqp.core.instruction.gates.gate import Gate
from mpqp.core.instruction.gates.gate_definition import GateDefinition

# 3M-TODO: there might be a conception problem: to initialize a gate we need a gate
#  definition, the easiest for a definition is to compute the matrix, the
#  computation of the matrix requires self.parameters which only exist after the
#  initialization of ParametrizedGate. For now this problem is manually tackled
#  by instantiating self.parameters before the computation of the gates's
#  definition but this solution looks like a conception problem...


@typechecked
class ParametrizedGate(Gate, ABC):
    """
    Define a parametrized gate.

    Args:
        definition: Provide a definition of the gate (matrix, gate combination, ...).
        targets: List of indices referring to the qubits on which the gate will
            be applied.
        parameters: List of parameters used to define the gate.
        label: Label used to identify the measurement.

    Example:
        >>> theta = np.pi/3
        >>> c, s = np.cos(theta / 2), np.sin(theta / 2)
        >>> gate_def = UnitaryMatrix(np.array([[c, s], [-s, c]]))
        >>> parametrized = ParametrizedGate(gate_def, 3, theta)
    """

    def __init__(
        self,
        definition: GateDefinition,
        targets: list[int],
        parameters: list[Expr | float],
        label: Optional[str] = None,
    ):
        Gate.__init__(self, targets, label)
        self.definition = definition
        """See parameter description."""
        self.parameters = parameters
        """See parameter description."""
        self._numeric_parameters = False

    def subs(
        self, values: dict[Expr | str, Complex], remove_symbolic: bool = False
    ) -> ParametrizedGate:
        concrete_gate = deepcopy(self)
        options = getattr(self, "native_gate_options", {})
        concrete_gate.definition = concrete_gate.definition.subs(
            values, remove_symbolic, **options
        )
        caster = lambda v: float(v) if remove_symbolic else v
        concrete_gate.parameters = [
            caster(param.subs(values)) if isinstance(param, Expr) else param
            for param in self.parameters
        ]
        concrete_gate._numeric_parameters = True

        return concrete_gate



# from getpass import getpass
# from unittest.mock import MagicMock, patch
#
# import pytest
# from qiskit_ibm_provider.api.exceptions import RequestsApiError
#
# from mpqp.execution.connection.ibm_connection import (
#     IBMProvider,
#     config_ibm_account,
#     setup_ibm_account,
#     test_connection,
# )
#
#
# @pytest.fixture
# def mock_ibm_provider(monkeypatch):
#     mock_ibm_provider = MagicMock()
#     monkeypatch.setattr(
#         "mpqp.execution.connection.ibm_connection.IBMProvider", mock_ibm_provider
#     )
#     return mock_ibm_provider
#
#
# @pytest.fixture
# def mock_getpass(monkeypatch):
#     mock_getpass = MagicMock(return_value="test_token")
#     monkeypatch.setattr("getpass.getpass", mock_getpass)
#     return mock_getpass
#
#
# def test_config_ibm_account():
#     token = "test_token"
#     mock_ibm_provider = MagicMock()
#
#     # successful configuration
#     with patch(
#         "mpqp.execution.connection.ibm_connection.IBMProvider", mock_ibm_provider
#     ):
#         config_ibm_account(token)
#
#     mock_ibm_provider.save_account.assert_called_once_with(token=token, overwrite=True)
#
#
# def test_setup_ibm_account_configured():
#     # test case when an IBMQ account is already configured
#     with patch(
#         "mpqp.execution.connection.env_manager.get_env_variable", return_value="True"
#     ), patch("builtins.input", side_effect=["n"]):
#         result, _ = setup_ibm_account()
#         assert result == "Canceled."
#
#
# def test_setup_ibm_account_not_configured():
#     # test case when an IBMQ account is not configured
#     with patch(
#         "mpqp.execution.connection.env_manager.get_env_variable", return_value=None
#     ), patch("builtins.input", side_effect=["y"]), patch(
#         "mpqp.execution.connection.ibm_connection.test_connection", return_value=True
#     ):
#         result, _ = setup_ibm_account()
#         assert result == "IBM Q account correctly configured"
#
#
# def test_setup_ibm_account_failed_and_not_configured():
#     # test case when an IBMQ account is not configured and test_connection fails
#     with patch(
#         "mpqp.execution.connection.env_manager.get_env_variable", return_value=None
#     ), patch("builtins.input", side_effect=["y"]), patch(
#         "mpqp.execution.connection.ibm_connection.test_connection", return_value=False
#     ):
#         result, _ = setup_ibm_account()
#         assert result == ""
#
#
# def test_test_connection():
#     if test_connection():
#         print("Test connection success - unexpected")
#
#     with patch(
#         "mpqp.execution.connection.ibm_connection.IBMProvider",
#         side_effect=RequestsApiError("Login failed"),
#     ):
#         try:
#             test_connection()
#         except RequestsApiError:
#             pass
#
#
# def test_mock_getpass(mock_getpass):
#     with patch("getpass.getpass", mock_getpass):
#         assert getpass() == "test_token"
#
#
# def test_mock_ibm_provider(mock_ibm_provider):
#     with patch(
#         "mpqp.execution.connection.ibm_connection.IBMProvider", mock_ibm_provider
#     ):
#         with pytest.raises(RequestsApiError, match="401 Client Error: Unauthorized"):
#             IBMProvider()



import pytest
from mpqp import QCircuit
from mpqp.core.instruction.measurement import Observable, ExpectationMeasure
from mpqp.execution.devices import AWSDevice, ATOSDevice, IBMDevice
from mpqp.gates import *
from mpqp.measures import BasisMeasure
from mpqp.execution import run
from braket.devices import LocalSimulator
import numpy as np
from mpqp.qasm.qasm_to_braket import qasm3_to_braket_Circuit


def test_sample_demo():
    # Declaration of the circuit with the right size
    circuit = QCircuit(4)

    # Constructing the circuit by adding gates
    circuit.add(T(0))
    circuit.add(CNOT(0, 1))
    circuit.add(X(0))
    circuit.add(H(1))
    circuit.add(Z(2))
    circuit.add(CZ(2, 1))
    circuit.add([SWAP(2, 0), CNOT(0, 2)])
    circuit.add(Ry(3.14 / 2, 2))
    circuit.add(S(1))
    circuit.add(H(3))
    circuit.add(CNOT(1, 2))
    circuit.add(Rx(3.14, 1))
    circuit.add(CNOT(3, 0))
    circuit.add(Rz(3.14, 0))

    # Add measurement
    circuit.add(BasisMeasure([0, 1, 2, 3], shots=2000))

    # Run the circuit on a selected device
    run(
        circuit,
        [
            IBMDevice.AER_SIMULATOR,
            ATOSDevice.MYQLM_PYLINALG,
            ATOSDevice.MYQLM_CLINALG,
            AWSDevice.BRAKET_LOCAL_SIMULATOR,
        ],
    )

    assert True


def test_statevector_demo():
    # Declaration of the circuit with the right size
    circuit = QCircuit(4)

    # Constructing the circuit by adding gates
    circuit.add(T(0))
    circuit.add(CNOT(0, 1))
    circuit.add(X(0))
    circuit.add(H(1))
    circuit.add(Z(2))
    circuit.add(CZ(2, 1))
    circuit.add([SWAP(2, 0), CNOT(0, 2)])
    circuit.add(Ry(3.14 / 2, 2))
    circuit.add(S(1))
    circuit.add(H(3))
    circuit.add(CNOT(1, 2))
    circuit.add(Rx(3.14, 1))
    circuit.add(CNOT(3, 0))
    circuit.add(Rz(3.14, 0))

    # when no measure in the circuit, must run in statevector mode
    run(
        circuit,
        [
            IBMDevice.AER_SIMULATOR_STATEVECTOR,
            ATOSDevice.MYQLM_PYLINALG,
            ATOSDevice.MYQLM_CLINALG,
            AWSDevice.BRAKET_LOCAL_SIMULATOR,
        ],
    )

    # same when we add a BasisMeasure with 0 shots
    circuit.add(BasisMeasure([0, 1, 2, 3], shots=0))

    # Run the circuit on a selected device
    run(
        circuit,
        [
            IBMDevice.AER_SIMULATOR_STATEVECTOR,
            ATOSDevice.MYQLM_PYLINALG,
            ATOSDevice.MYQLM_CLINALG,
            AWSDevice.BRAKET_LOCAL_SIMULATOR,
        ],
    )

    assert True


def test_observable_demo():
    obs = Observable(
        np.array(
            [
                [0.63, 0.5, 1, 1],
                [0.5, 0.82, 1, 1],
                [1, 1, 1, 0.33],
                [1, 1, 0.33, 0.3],
            ],
            dtype=float,
        )
    )

    # Declaration of the circuit with the right size
    circuit = QCircuit(2, label="Observable test")
    # Constructing the circuit by adding gates and measurements
    circuit.add(H(0))
    circuit.add(Rx(1.76, 1))
    circuit.add(ExpectationMeasure([0, 1], observable=obs, shots=0))

    # Running the computation on myQLM and on Aer simulator, then retrieving the results
    run(circuit, [ATOSDevice.MYQLM_PYLINALG, IBMDevice.AER_SIMULATOR])

    assert True


def test_aws_executions():
    device = LocalSimulator()

    qasm_str = """OPENQASM 3.0;
    include 'stdgates.inc';
    qubit[2] q;
    bit c;
    h q[0];
    cx q[0],q[1];
    c[0] = measure q[0];
    c[1] = measure q[1];"""

    circuit = qasm3_to_braket_Circuit(qasm_str)

    device.run(circuit, shots=100).result()

    #####################################################

    # Declaration of the circuit with the right size
    circuit = QCircuit(4)

    # Constructing the circuit by adding gates
    circuit.add(T(0))
    circuit.add(CNOT(0, 1))
    circuit.add(X(0))
    circuit.add(H(1))
    circuit.add(Z(2))
    circuit.add(CZ(2, 1))
    circuit.add(SWAP(2, 0))
    circuit.add(CNOT(0, 2))
    circuit.add(Ry(3.14 / 2, 2))
    circuit.add(S(1))
    circuit.add(H(3))
    circuit.add(CNOT(1, 2))
    circuit.add(Rx(3.14, 1))
    circuit.add(CNOT(3, 0))
    circuit.add(Rz(3.14, 0))
    # Add measurement
    circuit.add(BasisMeasure([0, 1, 2, 3], shots=2000))

    run(circuit, AWSDevice.BRAKET_LOCAL_SIMULATOR)

    #####################################################

    obs = Observable(
        np.array(
            [
                [0.63, 0.5, 1, 1],
                [0.5, 0.82, 1, 1],
                [1, 1, 1, 0.33],
                [1, 1, 0.33, 0.3],
            ],
            dtype=float,
        )
    )

    # Declaration of the circuit with the right size
    circuit = QCircuit(2, label="Observable test")
    # Constructing the circuit by adding gates and measurements
    circuit.add(H(0))
    circuit.add(Rx(1.76, 1))
    circuit.add(ExpectationMeasure([0, 1], observable=obs, shots=0))

    # Running the computation on myQLM and on Braket simulator, then retrieving the results
    run(circuit, [AWSDevice.BRAKET_LOCAL_SIMULATOR, ATOSDevice.MYQLM_PYLINALG])

    #####################################################

    # Declaration of the circuit with the right size
    circuit = QCircuit(
        [H(0), Rx(1.76, 1), Ry(1.76, 1), Rz(1.987, 0)], label="StateVector test"
    )

    # Running the computation on myQLM and on Aer simulator, then retrieving the results
    run(circuit, [AWSDevice.BRAKET_LOCAL_SIMULATOR, ATOSDevice.MYQLM_PYLINALG])


def test_all_native_gates():
    # Declaration of the circuit with the right size
    circuit = QCircuit(3, label="Test native gates")
    # Constructing the circuit by adding gates and measurements
    circuit.add([H(0), X(1), Y(2), Z(0), S(1), T(0)])
    circuit.add([Rx(1.2324, 2), Ry(-2.43, 0), Rz(1.04, 1), Rk(-1, 1), P(-323, 2)])
    circuit.add(U(1.2, 2.3, 3.4, 2))
    circuit.add(SWAP(2, 0))
    circuit.add([CNOT(0, 1), CRk(4, 2, 1), CZ(1, 2)])
    circuit.add(TOF([0, 1], 2))

    circuit.to_qasm2()
    with pytest.warns(
        UserWarning,
        match=r"There is a phase e\^\(i\(a\+c\)/2\) difference between U\(a,b,c\) gate in 2.0 and 3.0.",
    ):
        circuit.to_qasm3()
        run(
            circuit,
            [
                ATOSDevice.MYQLM_PYLINALG,
                ATOSDevice.MYQLM_CLINALG,
                IBMDevice.AER_SIMULATOR_STATEVECTOR,
                AWSDevice.BRAKET_LOCAL_SIMULATOR,
            ],
        )



from __future__ import annotations

from typing import Optional, Union

import numpy as np
from sympy import Expr
from typeguard import typechecked
from numbers import Complex

from mpqp.core.circuit import QCircuit
from mpqp.core.instruction.measurement.basis_measure import BasisMeasure
from mpqp.core.instruction.measurement.expectation_value import (
    ExpectationMeasure,
    Observable,
)
from mpqp.execution.devices import AvailableDevice, IBMDevice, ATOSDevice, AWSDevice
from mpqp.execution.providers_execution.aws_execution import (
    run_braket,
    submit_job_braket,
)
from mpqp.execution.providers_execution.atos_execution import run_atos, submit_QLM
from mpqp.execution.providers_execution.ibm_execution import run_ibm, submit_ibmq
from mpqp.execution.result import Result, BatchResult
from mpqp.execution.job import Job, JobType, JobStatus
from mpqp.tools.errors import RemoteExecutionError


@typechecked
def adjust_measure(measure: ExpectationMeasure, circuit: QCircuit):
    """We allow the measure to not span the entire circuit, but none of our
    providers allow for that. So we patch the measure with this function, with
    identities on the qubits that no not interest our user.

    Args:
        measure: The expectation measure, potentially incomplete.
        circuit: The circuit to which will be added the potential swaps allowing
            the user to get the expectation value of the qubits in an arbitrary
            order (this part is not handled by this function).

    Returns:
        The measure padded with the identities before and after.
    """
    Id_before = np.eye(2 ** measure.rearranged_targets[0])
    Id_after = np.eye(2 ** (circuit.nb_qubits - measure.rearranged_targets[-1] - 1))
    tweaked_measure = ExpectationMeasure(
        list(range(circuit.nb_qubits)),
        Observable(
            np.kron(
                np.kron(Id_before, measure.observable.matrix), Id_after
            )  # pyright: ignore[reportArgumentType]
        ),
        measure.shots,
    )
    return tweaked_measure


@typechecked
def generate_job(
    circuit: QCircuit, device: AvailableDevice, values: dict[Expr | str, Complex] = {}
) -> Job:
    """Create a Job from the circuit and the eventual measurements. If the
    circuit depends on variables, the values given in parameters are used to do
    the substitution.

    Args:
        circuit: Circuit to be run.
        device: Device on which the circuit will be run.
        values: Set of values to substitute symbolic variables.

    Returns:
        The Job containing information about the execution of the circuit.
    """
    circuit = circuit.subs(values, True)

    # get the measurements of this circuit
    m_list = circuit.get_measurements()
    nb_meas = len(m_list)

    # determine the job type and create the right job
    if nb_meas == 0:
        job = Job(JobType.STATE_VECTOR, circuit, device)
    elif nb_meas == 1:
        measurement = m_list[0]
        if isinstance(measurement, BasisMeasure):
            # 3M-TODO: handle other basis by adding the right rotation (change of basis) before
            #       measuring in the computational basis
            # 3M-TODO: Muhammad circuit.add(CustomGate(UnitaryMatrix(change_of_basis_inverse)))
            if measurement.shots <= 0:
                job = Job(JobType.STATE_VECTOR, circuit, device)
            else:
                job = Job(JobType.SAMPLE, circuit, device, measurement)
        elif isinstance(measurement, ExpectationMeasure):
            job = Job(
                JobType.OBSERVABLE,
                circuit + measurement.pre_measure,
                device,
                adjust_measure(measurement, circuit),
            )
        else:
            raise NotImplementedError(
                f"Measurement type {type(measurement)} not handled"
            )
    else:
        raise NotImplementedError(
            "Cannot handle several measurement in the current version"
        )

    return job


@typechecked
def _run_single(
    circuit: QCircuit, device: AvailableDevice, values: dict[Expr | str, Complex]
) -> Result:
    """Runs the circuit on the ``backend``. If the circuit depends on variables,
    the ``values`` given in parameters are used to do the substitution.

    Args:
        circuit: QCircuit to be run.
        device: Device, on which the circuit will be run.
        values: Set of values to substitute symbolic variables. Defaults to ``{}``.

    Returns:
        The Result containing information about the measurement required.

    Example:
        >>> c = QCircuit([H(0), CNOT(0, 1), BasisMeasure([0, 1], shots=1000)], label="Bell pair")
        >>> result = run(c, IBMDevice.AER_SIMULATOR)
        >>> print(result)
        Result: IBMDevice, AER_SIMULATOR
        Counts: [512, 0, 0, 488]
        Probabilities: [0.512 0.    0.    0.488]
        State: 00, Index: 0, Count: 512, Probability: 0.512
        State: 11, Index: 3, Count: 488, Probability: 0.488
        Error: None
    """
    job = generate_job(circuit, device, values)
    job.status = JobStatus.INIT

    if isinstance(device, IBMDevice):
        return run_ibm(job)
    elif isinstance(device, ATOSDevice):
        return run_atos(job)
    elif isinstance(device, AWSDevice):
        return run_braket(job)
    else:
        raise NotImplementedError(f"Device {device} not handled")


@typechecked
def run(
    circuit: QCircuit,
    device: AvailableDevice | list[AvailableDevice],
    values: Optional[dict[Expr | str, Complex]] = None,
) -> Union[Result, BatchResult]:
    """Runs the circuit on the backend, or list of backend, provided in
    parameter. If the circuit depends on variables, the values given in
    parameters are used to do the substitution.

    Args:
        circuit: QCircuit to be run.
        device: Device, or list of devices, on which the circuit will be run.
        values: Set of values to substitute symbolic variables. Defaults to ``{}``.

    Returns:
        The Result containing information about the measurement required.

    Examples:
        >>> c = QCircuit(
        ...     [H(0), CNOT(0, 1), BasisMeasure([0, 1], shots=1000)],
        ...     label="Bell pair",
        ... )
        >>> result = run(c, IBMDevice.AER_SIMULATOR)
        >>> print(result)
        Result: IBMDevice, AER_SIMULATOR
        Counts: [512, 0, 0, 488]
        Probabilities: [0.512 0.    0.    0.488]
        State: 00, Index: 0, Count: 512, Probability: 0.512
        State: 11, Index: 3, Count: 488, Probability: 0.488
        Error: None
        >>> batch_result = run(
        ...     c,
        ...     [ATOSDevice.MYQLM_PYLINALG, AWSDevice.BRAKET_LOCAL_SIMULATOR]
        ... )
        >>> print(batch_result)
        BatchResult: 2 results
        Result: AWSDevice, BRAKET_LOCAL_SIMULATOR
        Counts: [492, 0, 0, 508]
        Probabilities: [0.492 0.    0.    0.508]
        State: 00, Index: 0, Count: 492, Probability: 0.492
        State: 11, Index: 3, Count: 508, Probability: 0.508
        Error: None
        Result: ATOSDevice, MYQLM_PYLINALG
        Counts: [462, 0, 0, 538]
        Probabilities: [0.462 0.    0.    0.538]
        State: 00, Index: 0, Count: 462, Probability: 0.462
        State: 11, Index: 3, Count: 538, Probability: 0.538
        Error: 0.015773547629015002
    """

    if values is None:
        values = {}

    if isinstance(device, list):
        # Duplicate devices are removed
        set_device = list(set(device))
        if len(set_device) == 1:
            return _run_single(circuit, set_device[0], values)

        return BatchResult([_run_single(circuit, dev, values) for dev in set_device])

    return _run_single(circuit, device, values)


@typechecked
def submit(
    circuit: QCircuit, device: AvailableDevice, values: dict[Expr | str, Complex] = {}
) -> tuple[str, Job]:
    """
    Submit the job related with the circuit on the remote backend provided in parameter.
    The submission returns a job_id that can be used to retrieve the Result later.
    If the circuit depends on variables, the values given in parameters are used to do the substitution.
    Unlike :meth:`run`, for the moment, one can only submit a circuit to a single device.

    Example:
        >>> circuit = QCircuit([H(0), CNOT(0,1), BasisMeasure([0,1], shots=10)])
        >>> job_id, job = submit(circuit, ATOSDevice.QLM_LINALG)
        Logging as user <qlm_user>...
        Submitted a new batch: Job766
        >>> print("Status of " +job_id +":", job.job_status)
        Status of Job766: JobStatus.RUNNING

    Args:
        circuit: QCircuit to be run.
        device: Remote device on which the circuit will be submitted.
        values: Set of values to substitute symbolic variables.

    Returns:
        The job id provided by the remote device after submission of the job.
    """
    if not device.is_remote():
        raise RemoteExecutionError(
            "submit(...) function is only made for remote device."
        )

    job = generate_job(circuit, device, values)
    job.status = JobStatus.INIT

    if isinstance(device, IBMDevice):
        job_id, _ = submit_ibmq(job)
    elif isinstance(device, ATOSDevice):
        job_id, _ = submit_QLM(job)
    elif isinstance(device, AWSDevice):
        job_id, _ = submit_job_braket(job)
    else:
        raise NotImplementedError(f"Device {device} not handled")

    return job_id, job



"""The :class:`BasisMeasure` is used to project the state on a statevector of a
given :class:`Basis` and returns the corresponding eigenvalue."""

from __future__ import annotations
from typing import Optional

import qiskit.circuit
from qiskit.circuit import Parameter
from typeguard import typechecked

from .measure import Measure
from .basis import Basis, ComputationalBasis
from mpqp.core.languages import Language


@typechecked
class BasisMeasure(Measure):
    """Class representing a measure of one or several qubits in a specific
    basis.

    By default, the computational basis will be used. The user can also precise
    a specific basis using the class Basis.

    The number of shots indicates the number of time the measure is repeated.
    When shots is equal to 0 (by default), the simulator is used to produce
    exact value of the amplitudes/probabilities.

    Args:
        targets: List of indices referring to the qubits on which the measure
            will be applied.
        c_targets: List of indices referring to the classical bits on which the
            measure will be applied.
        shots: Number of shots to be performed basis: basis in which the qubits
            should be measured.
        basis: Basis in which the measure is performed. Defaults to
            :class:`ComputationalBasis()<mpqp.core.instruction.measurement.basis.ComputationalBasis>`
        label: Label used to identify the measure.

    Examples:
        >>> c1 = QCircuit([H(0), H(1), CNOT(0,1)])
        >>> c1.add(BasisMeasure([0, 1, 2], shots=1024))
        >>> c2 = QCircuit([H(0), H(1), CNOT(0,1)])
        >>> c2.add(BasisMeasure([0, 1, 2], shots=1024, basis=HadamardBasis()))
    """

    def __init__(
        self,
        targets: list[int],
        c_targets: Optional[list[int]] = None,
        shots: int = 0,
        basis: Optional[Basis] = None,
        label: Optional[str] = None,
    ):
        if basis is None:
            basis = ComputationalBasis()
        # 6M-TODO: implement basis thing
        if c_targets is not None:
            if len(set(c_targets)) != len(c_targets):
                raise ValueError(f"Duplicate registers in targets: {c_targets}")
        super().__init__(targets, shots, label)
        self.user_set_c_targets = c_targets is not None
        self.c_targets = c_targets
        """See parameter description."""
        self.basis = basis
        """See parameter description."""

    def to_other_language(
        self,
        language: Language = Language.QISKIT,
        qiskit_parameters: Optional[set[Parameter]] = None,
    ):
        if qiskit_parameters is None:
            qiskit_parameters = set()
        if language == Language.QISKIT:
            if isinstance(self.basis, ComputationalBasis):
                return qiskit.circuit.Measure()
            else:
                raise NotImplementedError(f"{type(self.basis)} not handled")
        else:
            raise NotImplementedError(f"{language} is not supported")

    def __repr__(self) -> str:
        options = ""
        if self.shots != 0:
            options += f", shots={self.shots}"
        if not isinstance(self.basis, ComputationalBasis):
            options += f", basis={self.basis}"
        if self.label is not None:
            options += f", label={self.label}"
        return f"BasisMeasure({self.targets}{options})"



import os
from typing import Any

from botocore.exceptions import NoRegionError
from braket.aws import AwsDevice, AwsSession
from braket.devices import LocalSimulator
from braket.devices.device import Device as BraketDevice
from termcolor import colored
from typeguard import typechecked

from mpqp.execution.connection.env_manager import get_env_variable, save_env_variable
from mpqp.execution.devices import AWSDevice
from mpqp.tools.errors import AWSBraketRemoteExecutionError


def setup_aws_braket_account() -> tuple[str, list[Any]]:
    """Setups the connection to an Amazon Braket account using user input.

    This function checks whether an Amazon Braket account is already configured
    and prompts the user to update it if needed. It then collects the user's AWS
    access key, AWS secret key (hidden input), and the AWS region for Amazon
    Braket. The function attempts to configure the Amazon Braket account using
    the provided credentials.

    Returns:
        A tuple containing a message indicating the result of the setup (e.g.,
        success, cancelled, or error, ...) and an empty list. The list is
        included for consistency with the existing code structure.
    """

    already_configured = get_env_variable("BRAKET_CONFIGURED") == "True"

    if already_configured:
        decision = input(
            "An Amazon Braket account is already configured. Do you want to update it? [y/N] "
        )
        if decision.lower().strip() != "y":
            return "Canceled.", []

    try:
        os.system("aws configure")
        save_env_variable("BRAKET_CONFIGURED", "True")
        return "Amazon Braket account correctly configured", []

    except Exception as e:
        print(colored("Error configuring Amazon Braket account", "red"))
        print(colored(str(e), "red"))
        input("Press 'Enter' to continue")
        return "", []


def get_aws_braket_account_info() -> str:
    """Get AWS Braket credentials information including access key ID,
    obfuscated secret access key, and region.

    Example:
        >>> get_aws_braket_account_info()
            access_key_id: 'AKIA26NYJ***********'
            secret_access_key: 'sMDad***********************************'
            region: 'us-east-1'

    Returns:
        A formatted string containing AWS credentials information with an
        obfuscated secret access key.

    """
    if get_env_variable("BRAKET_CONFIGURED") == "False":
        raise AWSBraketRemoteExecutionError(
            "Error when trying to get AWS credentials. No AWS Braket account configured."
        )

    try:
        session = AwsSession()

        # get the AWS Braket user access key, secret key and region
        access_key_id = session.boto_session.get_credentials().access_key

        secret_access_key = session.boto_session.get_credentials().secret_key
        obfuscate_key = secret_access_key[:5] + "*" * (len(secret_access_key) - 5)

        region_name = session.boto_session.region_name
    except Exception as e:
        raise AWSBraketRemoteExecutionError(
            "Error when trying to get AWS credentials. No AWS Braket account configured.\n Trace:"
            + str(e)
        )

    return f"""    access_key_id: '{access_key_id}'
    secret_access_key: '{obfuscate_key}' 
    region: '{region_name}'"""


@typechecked
def get_braket_device(device: AWSDevice) -> BraketDevice:
    """
    Returns the AwsDevice device associate with the AWSDevice in parameter.

    Example:
        >>> device = get_braket_device(AWSDevice.BRAKET_RIGETTI_ASPEN_M_3)
        >>> device.properties.action['braket.ir.jaqcd.program'].supportedResultTypes
        [ResultType(name='Sample', observables=['x', 'y', 'z', 'h', 'i'], minShots=10, maxShots=100000),
         ResultType(name='Expectation', observables=['x', 'y', 'z', 'h', 'i'], minShots=10, maxShots=100000),
         ResultType(name='Variance', observables=['x', 'y', 'z', 'h', 'i'], minShots=10, maxShots=100000),
         ResultType(name='Probability', observables=None, minShots=10, maxShots=100000)]

    Args:
        device: AWSDevice element describing which remote/local AwsDevice we want.

    Raises:
        AWSBraketRemoteExecutionError
    """

    if not device.is_remote():
        return LocalSimulator()

    try:
        return AwsDevice(device.get_arn())
    except ValueError as ve:
        raise AWSBraketRemoteExecutionError(
            "Failed to retrieve remote AWS device. Please check the arn, or if the "
            "device is accessible from your region..\nTrace: " + str(ve)
        )
    except NoRegionError as err:
        raise AWSBraketRemoteExecutionError(
            "Failed to find the region related with your aws credentials. Make sure you"
            "configured correctly your AWS account with 'setup_connections.py' script."
            "\nTrace: " + str(err)
        )


def get_all_task_ids() -> list[str]:
    """Retrieves all the task ids of this account/group from AWS.

    Example:
        >>> get_all_task_ids()
        ['arn:aws:braket:us-east-1:752542621531:quantum-task/6a46ae9a-d02f-4a23-b46f-eae43471bc22',
         'arn:aws:braket:us-east-1:752542621531:quantum-task/11db7e68-2b17-4b00-a4ec-20f662fd4876',
         'arn:aws:braket:us-east-1:752542621531:quantum-task/292d329f-727c-4b92-83e1-7d4bedd4b243',
         'arn:aws:braket:us-east-1:752542621531:quantum-task/4b94c703-2ce8-480b-b3f3-ecb2580dbb82',
         'arn:aws:braket:us-east-1:752542621531:quantum-task/edc094aa-23e8-4a8c-87be-f2e09281d79d',
         'arn:aws:braket:us-east-1:752542621531:quantum-task/af9e623a-dd1c-4ecb-9db6-dbbd1af08110']
    """
    return [
        task["quantumTaskArn"]
        for task in (
            AwsSession().braket_client.search_quantum_tasks(filters=[])["quantumTasks"]
        )
    ]


def get_all_partial_ids() -> list[str]:
    """Retrieves all the task ids of this account/group from AWS and extracts the
    significant part.

    Example:
        >>> get_all_partial_ids()
        ['6a46ae9a-d02f-4a23-b46f-eae43471bc22',
         '11db7e68-2b17-4b00-a4ec-20f662fd4876',
         '292d329f-727c-4b92-83e1-7d4bedd4b243',
         '4b94c703-2ce8-480b-b3f3-ecb2580dbb82',
         'edc094aa-23e8-4a8c-87be-f2e09281d79d',
         'af9e623a-dd1c-4ecb-9db6-dbbd1af08110']
    """
    return [id.split("/")[-1] for id in get_all_task_ids()]



import pytest

import numpy as np
import numpy.typing as npt

from mpqp.gates import Gate, Z, X, SWAP, CustomGate, UnitaryMatrix


@pytest.mark.parametrize(
    "gate, inverse",
    [
        (Z(0), Z(0)),
        (
            CustomGate(UnitaryMatrix(np.diag([1, 1j])), [0]),
            CustomGate(UnitaryMatrix(np.diag([1, -1j])), [0]),
        ),
    ],
)
def test_inverse(gate: Gate, inverse: Gate):
    assert gate.inverse().is_equivalent(inverse)


@pytest.mark.parametrize(
    "g1, g2",
    [
        (Z(0), Z(0)),
        (
            CustomGate(UnitaryMatrix(np.diag([1, -1])), [0]),
            Z(0),
        ),
        (
            CustomGate(UnitaryMatrix(np.diag([1.00000001, -1])), [0]),
            Z(0),
        ),
    ],
)
def test_is_equivalent(g1: Gate, g2: Gate):
    assert g1.is_equivalent(g2)


@pytest.mark.parametrize(
    "g1, g2",
    [
        (Z(0), X(0)),
        (
            CustomGate(UnitaryMatrix(np.diag([1, -1j])), [0]),
            Z(0),
        ),
    ],
)
def test_is_not_equivalent(g1: Gate, g2: Gate):
    assert not g1.is_equivalent(g2)


@pytest.mark.parametrize(
    "gate, pow, result_matrix",
    [
        (
            SWAP(0, 1),
            2,
            np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]]),
        ),
        (
            SWAP(0, 1),
            -1,
            np.array([[1, 0, 0, 0], [0, 0, 1, 0], [0, 1, 0, 0], [0, 0, 0, 1]]),
        ),
    ],
)
def test_power(gate: Gate, pow: float, result_matrix: npt.NDArray[np.complex64]):
    assert np.allclose(gate.power(pow).to_matrix(), result_matrix)


@pytest.mark.parametrize(
    "g1, g2, result_matrix",
    [
        (
            X(0),
            Z(0),
            np.array([[0, 0, 1, 0], [0, 0, 0, -1], [1, 0, 0, 0], [0, -1, 0, 0]]),
        ),
    ],
)
def test_tensor_product(g1: Gate, g2: Gate, result_matrix: npt.NDArray[np.complex64]):
    assert np.allclose(g1.tensor_product(g2).to_matrix(), result_matrix)


@pytest.mark.parametrize(
    "g1, g2, result_matrix",
    [
        (X(0), Z(0), np.array([[0, -1], [1, 0]])),
    ],
)
def test_product(g1: Gate, g2: Gate, result_matrix: npt.NDArray[np.complex64]):
    assert np.allclose(g1.product(g2).to_matrix(), result_matrix)


@pytest.mark.parametrize(
    "g1, scalar, result_matrix",
    [
        (X(0), 1j, np.array([[0, 1j], [1j, 0]])),
    ],
)
def test_scalar_product(
    g1: Gate, scalar: complex, result_matrix: npt.NDArray[np.complex64]
):
    assert np.allclose(g1.scalar_product(scalar).to_matrix(), result_matrix)


@pytest.mark.parametrize(
    "g1, g2, result_matrix",
    [
        (X(0), Z(0), np.array([[1, 1], [1, -1]]) / np.sqrt(2)),
    ],
)
def test_add_product(g1: Gate, g2: Gate, result_matrix: npt.NDArray[np.complex64]):
    assert np.allclose(g1.plus(g2).to_matrix(), result_matrix)


@pytest.mark.parametrize(
    "g1, g2, result_matrix",
    [
        (X(0), Z(0), np.array([[-1, 1], [1, 1]]) / np.sqrt(2)),
    ],
)
def test_sub_product(g1: Gate, g2: Gate, result_matrix: npt.NDArray[np.complex64]):
    assert np.allclose(g1.minus(g2).to_matrix(), result_matrix)



"""add -l or --long to the cli args to run this test (disabled by default 
because too slow)"""

# 3M-TODO test everything
import numpy as np
import pytest

from mpqp.core.instruction.measurement import Observable, ExpectationMeasure
from mpqp.gates import *
from mpqp import QCircuit
from mpqp.measures import BasisMeasure
from mpqp.execution import run
from mpqp.execution.devices import ATOSDevice

import sys


@pytest.mark.parametrize(
    "circuit",
    [
        # SAMPLE JOB
        QCircuit(
            [
                T(0),
                CNOT(0, 1),
                Ry(np.pi / 2, 2),
                S(1),
                CZ(2, 1),
                SWAP(2, 0),
                BasisMeasure(list(range(3)), shots=2000),
            ]
        ),
        # STATEVECTOR JOB
        QCircuit(
            [
                T(0),
                CNOT(0, 1),
                Ry(np.pi / 2, 2),
                S(1),
                CZ(2, 1),
                SWAP(2, 0),
                BasisMeasure(list(range(3)), shots=0),
            ]
        ),
        # OBSERVABLE JOB
        QCircuit(
            [
                H(0),
                Rx(1.76, 1),
                ExpectationMeasure(
                    [0, 1],
                    observable=Observable(
                        np.array(
                            [
                                [0.63, 0.5, 1, 1],
                                [0.5, 0.82, 1, 1],
                                [1, 1, 1, 0.33],
                                [1, 1, 0.33, 0.3],
                            ],
                        )
                    ),
                    shots=1000,
                ),
            ]
        ),
    ],
)
def running_remote_QLM_without_error(circuit: QCircuit):
    run(circuit, ATOSDevice.QLM_LINALG)


if "-l" in sys.argv or "--long" in sys.argv:
    test_running_remote_QLM_without_error = running_remote_QLM_without_error



"""File regrouping all features for translating QASM code to myQLM objects """
from qat.core.wrappers.circuit import Circuit
from qat.interop.openqasm import OqasmParser  # type: ignore
from typeguard import typechecked

from mpqp.qasm.open_qasm_2_and_3 import open_qasm_hard_includes


@typechecked
def qasm2_to_myqlm_Circuit(qasm_str: str) -> Circuit:
    """
    Converting a OpenQASM 2.0 code into a QLM Circuit.

    Args:
        qasm_str: A string representing the OpenQASM 2.0 code.

    Returns:
        A Circuit equivalent to the QASM code in parameter.
    """
    parser = OqasmParser(gates={"p": "PH", "u": "U"})  # requires myqlm-interop-1.9.3
    circuit = parser.compile(open_qasm_hard_includes(qasm_str, set()))
    return circuit



import pytest

from mpqp.qasm.qasm_to_qiskit import qasm2_to_QuantumCircuit
from qiskit import QuantumCircuit


@pytest.mark.parametrize(
    "qasm_code, gate_names",
    [
        (
            """OPENQASM 2.0;""",
            [],
        ),
        (
            """OPENQASM 2.0;
            include "qelib1.inc";

            qreg q[2];
            creg c[2];
            h q[0];
            cx q[0],q[1];

            measure q[0] -> c[0];
            measure q[1] -> c[1];""",
            [
                "h",
                "cx",
            ],
        ),
    ],
)
def test_qasm2_to_QuantumCircuit(qasm_code: str, gate_names: list[str]):
    circ = qasm2_to_QuantumCircuit(qasm_code)
    assert isinstance(circ, QuantumCircuit)
    for instr, expected_gate in zip(circ.data, gate_names):
        assert instr.operation.name == expected_gate



import pytest

from mpqp.qasm.qasm_to_myqlm import qasm2_to_myqlm_Circuit
from qat.core.wrappers.circuit import Circuit


@pytest.mark.parametrize(
    "qasm_code, gate_names",
    [
        (
            """OPENQASM 3.0;""",
            [],
        ),
        (
            """OPENQASM 2.0;

            qreg q[2];
            creg c[2];
            h q[0];
            cx q[0],q[1];

            measure q[0] -> c[0];
            measure q[1] -> c[1];""",
            ["H", "CNOT"],
        ),
    ],
)
def test_qasm2_to_myqlm_Circuit(qasm_code: str, gate_names: list[str]):
    circ = qasm2_to_myqlm_Circuit(qasm_code)
    assert isinstance(circ, Circuit)
    for op, expected_gate in zip(circ.ops, gate_names):
        assert op.gate == expected_gate



import pytest
import numpy as np

from mpqp.tools.generics import Matrix
from mpqp.gates import symbols
from mpqp.tools.maths import is_hermitian

x = symbols("x", real=True)


@pytest.mark.parametrize(
    "matrix, isHermitian", [
        (np.array([[1, 2j, 3j], [-2j, 4, 5j], [-3j, -5j, 6]]), True),
        (np.diag([1, 2, 3, 4]), True),
        (np.array([[1, 2, 3], [2, 4, 5], [3, 5, 6]]), True),
        (np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]), False),
        (np.diag([1, x]), True),
        (np.array([[1, x], [-x, 2]]), False)
    ]
)
def test_is_hermitian(matrix: Matrix, isHermitian: bool):
    assert is_hermitian(matrix) == isHermitian



from __future__ import annotations

from numpy import pi
import pytest

from mpqp import QCircuit
from mpqp.gates import *
from mpqp.measures import BasisMeasure
from mpqp.execution import Result, StateVector, Sample, Job, JobType
from mpqp.execution.devices import IBMDevice


@pytest.mark.parametrize(
    "job_type, data",
    [
        (JobType.STATE_VECTOR, 0.4),
        (JobType.STATE_VECTOR, []),
        (JobType.OBSERVABLE, []),
        (JobType.OBSERVABLE, StateVector([1])),
        (JobType.SAMPLE, 0.4),
        (JobType.SAMPLE, StateVector([1])),
    ],
)
def test_result_wrong_type(job_type: JobType, data: float | StateVector | list[Sample]):
    c = QCircuit(3)
    c.add(Rx(pi / 2, 1))
    j = Job(job_type, c, IBMDevice.AER_SIMULATOR)
    with pytest.raises(TypeError):
        Result(j, data)


@pytest.mark.parametrize(
    "job_type, data",
    [
        (JobType.STATE_VECTOR, StateVector([0.5, 0.5, 0.5, 0.5])),
        (JobType.OBSERVABLE, 0.4),
        (
            JobType.SAMPLE,
            [Sample(3, index=3, count=250), Sample(3, index=6, count=250)],
        ),
    ],
)
def test_result_right_type(job_type: JobType, data: float | StateVector | list[Sample]):
    size = 3
    c = QCircuit(size)
    c.add(Rx(pi / 2, 1))
    measure = BasisMeasure(list(range(size))) if job_type == JobType.SAMPLE else None
    j = Job(job_type, c, IBMDevice.AER_SIMULATOR, measure)
    if job_type == JobType.SAMPLE:
        assert isinstance(data, list)
        assert all(sample.count is not None for sample in data)
        shots = sum(sample.count for sample in data if sample.count is not None)
    else:
        shots = 0
    r = Result(j, data, shots=shots)
    assert r.device == r.job.device



from __future__ import annotations

from typing import Iterable, Optional, Sequence, Type, Union
from copy import deepcopy
from numbers import Complex

import numpy as np
import numpy.typing as npt
from matplotlib.figure import Figure
from qiskit.circuit import QuantumCircuit, Operation
from qiskit.circuit.quantumcircuit import CircuitInstruction
from qat.core.wrappers.circuit import Circuit as myQLM_Circuit
from braket.circuits import Circuit as braket_Circuit
from sympy import Basic, Expr
from typeguard import typechecked, TypeCheckError

from mpqp.core.instruction import Instruction
from mpqp.core.instruction.barrier import Barrier
from mpqp.core.instruction.gates import (
    ControlledGate,
    Gate,
)
from mpqp.core.instruction.gates.custom_gate import CustomGate
from mpqp.core.instruction.gates.gate_definition import UnitaryMatrix
from mpqp.core.instruction.gates.parametrized_gate import ParametrizedGate
from mpqp.core.instruction.measurement import (
    ComputationalBasis,
    BasisMeasure,
    Measure,
)
from mpqp.core.instruction.measurement.expectation_value import ExpectationMeasure
from mpqp.qasm import qasm2_to_myqlm_Circuit
from mpqp.qasm.open_qasm_2_and_3 import open_qasm_2_to_3
from mpqp.qasm.qasm_to_braket import qasm3_to_braket_Circuit
from mpqp.tools.errors import NumberQubitsError
from mpqp.core.languages import Language
from mpqp.tools.maths import matrix_eq


@typechecked
class QCircuit:
    """This class models a quantum circuit.

    A circuit is composed of quantum and classical bits and it is defined as a
    list of instructions applied on specific quantum and/or classical bits.

    Example:
        >>> circuit = QCircuit(2)
        >>> circuit.pretty_print()
        QCircuit : Size (Qubits,Cbits) = (2, 0), Nb instructions = 0
        q_0:
        q_1:
        >>> circuit = QCircuit(5, nb_cbits=2, label="Circuit 1")
        >>> circuit.add(Rx(1.23, 3))
        >>> circuit.pretty_print()
        QCircuit Circuit 1 : Size (Qubits,Cbits) = (5, 2), Nb instructions = 1
        q_0: ────────────
        q_1: ────────────
        q_2: ────────────
             ┌──────────┐
        q_3: ┤ Rx(1.23) ├
             └──────────┘
        q_4: ────────────

    Args:
        data: Number of qubits or List of instructions to initiate the circuit with. If the number of qubits is passed,
            it should be a positive int.
        nb_qubits: Optional number of qubits, in case you input the sequence of instruction and want to hardcode the
            number of qubits.
        nb_cbits: Number of classical bits. It should be positive. Defaults to None.
        label: Name of the circuit. Defaults to None.
    """

    def __init__(
        self,
        data: int | Sequence[Instruction],
        *,
        nb_qubits: Optional[int] = None,
        nb_cbits: Optional[int] = None,
        label: Optional[str] = None,
    ):
        self.nb_cbits = nb_cbits
        """See parameter description."""
        self.label = label
        """See parameter description."""
        self.instructions: list[Instruction] = []
        """List of instructions of the circuit."""
        self.nb_qubits: int
        """Number of qubits of the circuit."""

        if isinstance(data, int):
            if data < 0:
                raise TypeCheckError(
                    f"The data passed to QCircuit is a negative int ({data}), "
                    "this does not make sense."
                )
            self.nb_qubits = data
        else:
            if nb_qubits is None:
                if len(data) == 0:
                    self.nb_qubits = 0
                else:
                    connections: set[int] = set.union(
                        *(inst.connections() for inst in data)
                    )
                    self.nb_qubits = max(connections) + 1
            else:
                self.nb_qubits = nb_qubits
            self.add(map(deepcopy, data))

    def add(self, instruction: Instruction | Iterable[Instruction]):
        """Adds one instruction or a list of instructions at the end of the
        circuit.

        Args:
            instruction : Instruction(s) to append at the end of the circuit.

        Example:
            >>> circuit = QCircuit(2)
            >>> circuit.add(X(0))
            >>> circuit.add([CNOT(0, 1), BasisMeasure([0, 1], shots=100)])
            >>> circuit.pretty_print()
            QCircuit : Size (Qubits,Cbits) = (2, 2), Nb instructions = 3
                 ┌───┐     ┌─┐
            q_0: ┤ X ├──■──┤M├───
                 └───┘┌─┴─┐└╥┘┌─┐
            q_1: ─────┤ X ├─╫─┤M├
                      └───┘ ║ └╥┘
            c: 2/═══════════╩══╩═
                            0  1
        """
        if isinstance(instruction, Iterable):
            for inst in instruction:
                self.add(inst)
            return

        if any(conn > self.nb_qubits for conn in instruction.connections()):
            raise NumberQubitsError(
                f"Instruction {type(instruction)}'s connections "
                f"({instruction.connections()}) are not compatible with circuit"
                f" size ({self.nb_qubits})."
            )
        if any(qb >= self.nb_qubits for qb in instruction.targets):
            raise NumberQubitsError("Instruction targets qubit outside of circuit")
        if isinstance(instruction, ControlledGate):
            if any(qb >= self.nb_qubits for qb in instruction.controls):
                raise NumberQubitsError("Control targets qubit outside of circuit")

        if isinstance(instruction, BasisMeasure) and instruction.c_targets is None:
            if self.nb_cbits is None:
                self.nb_cbits = 0
            instruction.c_targets = [
                self.nb_cbits + i for i in range(len(instruction.targets))
            ]
            self.nb_cbits += len(instruction.c_targets)

        if isinstance(instruction, Barrier):
            instruction.size = self.nb_qubits

        self.instructions.append(instruction)

    def append(self, other: QCircuit, qubits_offset: int = 0) -> None:
        """Appends the circuit at the end (right side) of this circuit, inplace.

        If the size of the ``other`` is smaller than this circuit,
        the parameter ``qubits_offset`` can be used to indicate at which qubit
        the ``other`` circuit must be added.

        Example:
            >>> c1 = QCircuit([CNOT(0,1),CNOT(1,2)])
            >>> c2 = QCircuit([X(1),CNOT(1,2)])
            >>> c1.append(c2)
            >>> print(c1)
            q_0: ──■─────────────────
                 ┌─┴─┐     ┌───┐
            q_1: ┤ X ├──■──┤ X ├──■──
                 └───┘┌─┴─┐└───┘┌─┴─┐
            q_2: ─────┤ X ├─────┤ X ├
                      └───┘     └───┘

        Args:
            other: The circuit to append at the end of this circuit.
            qubits_offset: If the circuit in parameter is smaller, this
                parameter precise at which qubit (vertically) the circuit will
                be added.

        Raises:
            NumberQubitsError: if the circuit in parameter is larger than this
                circuit or if the ``qubits_offset`` is too big such that the
                ``other`` circuit would "stick out".
        """

        if self.nb_qubits < other.nb_qubits:
            raise NumberQubitsError(
                "Size of the circuit to be appended is greater than the size of"
                " this circuit"
            )
        if qubits_offset + other.nb_qubits > self.nb_qubits:
            raise NumberQubitsError(
                "Size of the circuit to be appended is too large given the"
                " index and the size of this circuit"
            )

        for inst in deepcopy(other.instructions):
            inst.targets = [qubit + qubits_offset for qubit in inst.targets]
            if isinstance(inst, ControlledGate):
                inst.controls = [qubit + qubits_offset for qubit in inst.controls]
            if isinstance(inst, BasisMeasure):
                if not inst.user_set_c_targets:
                    inst.c_targets = None

            self.add(inst)

    def __iadd__(self, other: QCircuit):
        self.append(other)
        return self

    def __add__(self, other: QCircuit) -> QCircuit:
        res = deepcopy(self)
        res += other
        return res

    def tensor(self, other: QCircuit) -> QCircuit:
        """Computes the tensor product of this circuit with the one in parameter.

        In the circuit notation, the upper part of the output circuit will correspond
        to the first circuit, while the bottom part correspond to the one in
        parameter.

        Example:
            >>> c1 = QCircuit([CNOT(0,1),CNOT(1,2)])
            >>> c2 = QCircuit([X(1),CNOT(1,2)])
            >>> print(c1.tensor(c2))
            q_0: ──■───────
                 ┌─┴─┐
            q_1: ┤ X ├──■──
                 └───┘┌─┴─┐
            q_2: ─────┤ X ├
                 ┌───┐└───┘
            q_3: ┤ X ├──■──
                 └───┘┌─┴─┐
            q_4: ─────┤ X ├
                      └───┘

        Args:
            other: QCircuit being the second operand of the tensor product with this circuit.

        Returns:
            The QCircuit resulting from the tensor product of this circuit with the one in parameter.
        """
        res = deepcopy(self)
        res.nb_qubits += other.nb_qubits
        res.append(other, qubits_offset=self.nb_qubits)
        return res

    def __matmul__(self, other: QCircuit) -> QCircuit:
        return self.tensor(other)

    def display(self, output: str = "mpl"):
        r"""Displays this circuit in the desired output format.

        For now, this uses the qiskit circuit drawer, so all formats supported
        by qiskit are supported.

        Example:
            >>> theta = symbols("θ")
            >>> circ = QCircuit([
            ...     P(theta, 0),
            ...     ExpectationMeasure([0], Observable(np.array([[0, 1], [1, 0]])), shots=1000)
            ... ])
            >>> circ.display("text")
               ┌──────┐
            q: ┤ P(θ) ├
               └──────┘
            >>> circ.display("latex")
            \documentclass[border=2px]{standalone}
            \usepackage[braket, qm]{qcircuit}
            \usepackage{graphicx}
            \begin{document}
            \scalebox{1.0}{
            \Qcircuit @C=1.0em @R=0.2em @!R { \\
                \nghost{{q} :  } & \lstick{{q} :  } & \gate{\mathrm{P}\,(\mathrm{{\ensuremath{\theta}}})} & \qw & \qw\\
            \\ }}
            \end{document}
            
        Args:
            output: Format of the output, see
                `docs.quantum.ibm.com/build/circuit-visualization <https://docs.quantum.ibm.com/build/circuit-visualization#alternative-renderers>`_
                for more information.
        """
        from qiskit.tools.visualization import circuit_drawer

        qc = self.to_other_language(language=Language.QISKIT)
        fig = circuit_drawer(qc, output=output, style={"backgroundcolor": "#EEEEEE"})
        if isinstance(fig, Figure):
            fig.show()
        return fig

    def size(self) -> tuple[int, int]:
        """Provides the size of the circuit, in terms of number of quantum and
        classical bits.

        Examples:
            >>> c1 = QCircuit([CNOT(0,1),CNOT(1,2)])
            >>> c1.size()
            (3, 0)
            >>> c2 = QCircuit(3,nb_cbits=2)
            >>> c2.size()
            (3, 2)
            >>> c3 = QCircuit([CNOT(0,1),CNOT(1,2), BasisMeasure([0,1,2], shots=200)])
            >>> c3.size()
            (3, 3)

        Returns:
            A couple ``(q, c)`` of integers, with ``q`` the number of qubits,
            and ``c`` the number of cbits of this circuit.
        """
        return self.nb_qubits, (self.nb_cbits or 0)

    def depth(self) -> int:
        """Computes the depth of the circuit.

        Example:
            >>> QCircuit([CNOT(0, 1), CNOT(1, 2), CNOT(0, 1), X(2)]).depth()
            3
            >>> QCircuit([CNOT(0, 1), CNOT(1, 2), CNOT(0, 1), Barrier(), X(2)]).depth()
            4

        Returns:
            Depth of the circuit.
        """
        if len(self) == 0:
            return 0

        nb_qubits = self.nb_qubits
        instructions = self.without_measurements().instructions
        layers = np.zeros((nb_qubits, self.count_gates()), dtype=bool)

        current_layer = 0
        last_barrier = 0
        for instr in instructions:
            if isinstance(instr, Barrier):
                last_barrier = current_layer
                current_layer += 1
                continue
            conns = list(instr.connections())
            if any(layers[conns, current_layer]):
                current_layer += 1
            fitting_layer_index = current_layer
            for index in range(current_layer, last_barrier - 1, -1):
                if any(layers[conns, index]):
                    fitting_layer_index = index + 1
                    break
            layers[conns, fitting_layer_index] = [True] * len(conns)

        return current_layer + 1

    def __len__(self) -> int:
        """Returns the number of instructions added to this circuit.

        Example:
            >>> c1 = QCircuit([CNOT(0,1), CNOT(1,2), X(1), CNOT(1,2)])
            >>> len(c1)
            4

        Returns:
            An integer representing the number of instructions in this circuit.
        """
        return len(self.instructions)

    def is_equivalent(self, circuit: QCircuit) -> bool:
        """Whether the circuit in parameter is equivalent to this circuit, in
        terms of gates, but not measurements.

        Depending on the definition of the gates of the circuit, several methods
        could be used to do it in an optimized way.

        Example:
            >>> c1 = QCircuit([H(0), H(0)])
            >>> c2 = QCircuit([Rx(0, 0)])
            >>> c1.is_equivalent(c2)
            True

        Args:
            circuit: The circuit for which we want to know if it is equivalent
                to this circuit.

        Returns:
            ``True`` if the circuit in parameter is equivalent to this circuit

        3M-TODO: will only work once the circuit.to_matrix is implemented
        """
        return matrix_eq(self.to_matrix(), circuit.to_matrix())

    def optimize(self, criteria: Optional[str | list[str]] = None) -> QCircuit:
        """Optimize the circuit to satisfy some criteria (depth, number of
        qubits, gate restriction) in parameter.

        Examples:
            >>>
            >>>
            >>>

        Args:
            criteria: String, or list of strings, regrouping the criteria of optimization of the circuit.

        Returns:
            the optimized QCircuit
        # 6M-TODO implement, example and test, can be an internship
        """
        # ideas: a circuit can be optimized
        # - to reduce the depth of the circuit (combine gates, simplify some sequences)
        # - according to a given topology or qubits connectivity map
        # - to avoid the use of some gates (imperfect or more noisy)
        # - to avoid multi-qubit gates
        ...

    def to_matrix(self) -> npt.NDArray[np.complex64]:
        """
        Compute the unitary matrix associated to this circuit.

        Examples:
            >>> c = QCircuit([H(0), CNOT(0,1)])
            >>> c.to_matrix()
            array([[ 0.70710678,  0.        ,  0.70710678,  0.        ],
                   [ 0.        ,  0.70710678,  0.        ,  0.70710678],
                   [ 0.        ,  0.70710678,  0.        , -0.70710678],
                   [ 0.70710678,  0.        , -0.70710678,  0.        ]])
            >>>

        Returns:
            a unitary matrix representing this circuit

        # 3M-TODO implement and double check examples and test:
        the idea is to compute the tensor product of the matrices associated
        with the gates of the circuit in a clever way (to minimize the number of
        multiplications) and then return the big matrix
        """
        ...

    def inverse(self) -> QCircuit:
        """Generate the inverse (dagger) of this circuit.

        Examples:
            >>> c1 = QCircuit([H(0), CNOT(0,1)])
            >>> print(c1)
                 ┌───┐
            q_0: ┤ H ├──■──
                 └───┘┌─┴─┐
            q_1: ─────┤ X ├
                      └───┘
            >>> print(c1.inverse())
                      ┌───┐
            q_0: ──■──┤ H ├
                 ┌─┴─┐└───┘
            q_1: ┤ X ├─────
                 └───┘
            >>> c2 = QCircuit([S(0), CZ(0,1), H(1), Ry(4.56, 1)])
            >>> print(c2)
                 ┌───┐
            q_0: ┤ S ├─■──────────────────
                 └───┘ │ ┌───┐┌──────────┐
            q_1: ──────■─┤ H ├┤ Ry(4.56) ├
                         └───┘└──────────┘
            >>> print(c2.inverse())

        Returns:
            The inverse circuit.

        # 3M-TODO implement, test, fill second example
        The inverse could be computed in several ways, depending on the
        definition of the circuit. One can inverse each gate in the circuit, or
        take the global unitary of the gate and inverse it.
        """
        dagger = QCircuit(self.nb_qubits)
        return dagger

    def to_gate(self) -> Gate:
        """Generate a gate from this entire circuit.

        Examples:
            >>> c = QCircuit([CNOT(0, 1), CNOT(1, 2), CNOT(0, 1), CNOT(2, 3)])
            >>> c.to_gate().definition.matrix
            >>>

        Returns:
            A gate representing this circuit.

        # 3M-TODO check implementation, example and test, this will only work
        chen circuit.to_matrix will be implemented
        """
        gate_def = UnitaryMatrix(self.to_matrix())
        return CustomGate(gate_def, list(range(self.nb_qubits)), label=self.label)

    @classmethod
    def initializer(cls, state: npt.NDArray[np.complex64]) -> QCircuit:
        """Initialize this circuit at a given state, given in parameter.

        This will imply adding gates at the beginning of the circuit.

        Examples:
            >>> qc = QCircuit.initializer(np.array([1, 0, 0 ,1])/np.sqrt(2))
            >>> print(qc)
                 ┌───┐
            q_0: ┤ H ├──■──
                 └───┘┌─┴─┐
            q_1: ─────┤ X ├
                      └───┘

        Args:
            state: StateVector modeling the state for initializing the circuit.

        Returns:
            A copy of the input circuit with additional instructions added
            before-hand to generate the right initial state.

        # 3M-TODO : to implement --> a first sort term way could be to reuse the
        # qiskit QuantumCircuit feature qc.initialize()
        """
        size = int(np.log2(len(state)))
        if 2**size != len(state):
            raise ValueError(f"Input state {state} should have a power of 2 size")
        res = cls(size)
        ...
        return res

    def count_gates(self, gate: Optional[Type[Gate]] = None) -> int:
        """Returns the number of gates contained in the circuit. If a specific
        gate is given in the ``gate`` arg, it returns the number of occurrences
        of this gate.

        Examples:
            >>> circuit = QCircuit(
            ...     [X(0), Y(1), Z(2), CNOT(0, 1), SWAP(0, 1), CZ(1, 2), X(2), X(1), X(0)]
            ... )
            >>> circuit.count_gates()
            9
            >>> circuit.count_gates(X)
            4
            >>> circuit.count_gates(Ry)
            0

        Args:
            gate: The gate for which we want to know its occurrence in this
                circuit.

        Returns:
            The number of gates (eventually of a specific type) contained in the
            circuit.
        """
        filter2 = Gate if gate is None else gate
        return len([inst for inst in self.instructions if isinstance(inst, filter2)])

    def get_measurements(self) -> list[Measure]:
        """Returns all the measurements present in this circuit.

        Example:
            >>> circuit = QCircuit([
            ...     BasisMeasure([0, 1], shots=1000),
            ...     ExpectationMeasure([1], Observable(np.identity(2)), shots=1000)
            ... ])
            >>> circuit.get_measurements()
            [BasisMeasure([0, 1], shots=1000),
             ExpectationMeasure([1], Observable(array([[1., 0.], [0., 1.]])), shots=1000)]

        Returns:
            The list of all measurements present in the circuit.
        """
        return [inst for inst in self.instructions if isinstance(inst, Measure)]

    def without_measurements(self) -> QCircuit:
        """Provides a copy of this circuit with all the measurements removed.

        Example:
            >>> circuit = QCircuit([X(0), CNOT(0, 1), BasisMeasure([0, 1], shots=100)])
            >>> print(circuit)
                 ┌───┐     ┌─┐
            q_0: ┤ X ├──■──┤M├───
                 └───┘┌─┴─┐└╥┘┌─┐
            q_1: ─────┤ X ├─╫─┤M├
                      └───┘ ║ └╥┘
            c: 2/═══════════╩══╩═
                            0  1
            >>> print(circuit.without_measurements())
                 ┌───┐
            q_0: ┤ X ├──■──
                 └───┘┌─┴─┐
            q_1: ─────┤ X ├
                      └───┘

        Returns:
            A copy of this circuit with all the measurements removed.
        """
        new_circuit = QCircuit(self.nb_qubits)
        new_circuit.instructions = [
            inst for inst in self.instructions if not isinstance(inst, Measure)
        ]

        return new_circuit

    def to_other_language(
        self, language: Language = Language.QISKIT
    ) -> Union[QuantumCircuit, myQLM_Circuit, braket_Circuit]:
        """Transforms this circuit into the corresponding circuit in the language
        specified in the ``language`` arg.

        By default, the circuit is translated to the corresponding
        ``QuantumCircuit`` in Qiskit, since it is the interface we use to
        generate the OpenQASM code.

        In the future, we will generate the OpenQASM code on our own, and this
        method will be used only for complex objects that are not tractable by
        OpenQASM (like hybrid structures).

        Example:
            >>> circuit = QCircuit([X(0), CNOT(0, 1)])
            >>> qc = circuit.to_other_language()
            >>> type(qc)
            qiskit.circuit.quantumcircuit.QuantumCircuit

        Args:
            language: Enum representing the target language.

        Returns:
            The corresponding circuit in the target language.
        """

        if language == Language.QISKIT:
            # to avoid defining twice the same parameter, we keep trace of the
            # added parameters, and we use those instead of new ones when they
            # are used more than once
            qiskit_parameters = set()
            if self.nb_cbits is None:
                new_circ = QuantumCircuit(self.nb_qubits)
            else:
                new_circ = QuantumCircuit(self.nb_qubits, self.nb_cbits)

            for instruction in self.instructions:
                if isinstance(instruction, ExpectationMeasure):
                    # these measures have no equivalent in Qiskit
                    continue
                qiskit_inst = instruction.to_other_language(
                    Language.QISKIT, qiskit_parameters
                )
                assert isinstance(qiskit_inst, CircuitInstruction) or isinstance(
                    qiskit_inst, Operation
                )
                cargs = []

                if isinstance(instruction, ControlledGate):
                    qargs = instruction.controls + instruction.targets
                elif isinstance(instruction, Gate):
                    qargs = instruction.targets
                elif isinstance(instruction, BasisMeasure) and isinstance(
                    instruction.basis, ComputationalBasis
                ):
                    assert instruction.c_targets is not None
                    qargs = [instruction.targets]
                    cargs = [instruction.c_targets]
                elif isinstance(instruction, Barrier):
                    qargs = range(instruction.size)
                else:
                    raise ValueError(f"Instruction not handled: {instruction}")

                new_circ.append(
                    qiskit_inst,
                    qargs,
                    cargs,
                )
            return new_circ

        elif language == Language.MY_QLM:
            cleaned_circuit = self.without_measurements()
            myqlm_circuit = qasm2_to_myqlm_Circuit(cleaned_circuit.to_qasm2())
            return myqlm_circuit

        elif language == Language.BRAKET:
            circuit_qasm3 = self.to_qasm3()
            brkt_circuit = qasm3_to_braket_Circuit(circuit_qasm3)
            return brkt_circuit

        else:
            raise NotImplementedError(f"Error: {language} is not supported")

    def to_qasm2(self) -> str:
        """Converts this circuit to the corresponding OpenQASM 2 code.

        For now, we use an intermediate conversion to a Qiskit
        ``QuantumCircuit``.

        Example:
            >>> circuit = QCircuit([X(0), CNOT(0, 1), BasisMeasure([0, 1], shots=100)])
            >>> print(circuit.to_qasm2())
            OPENQASM 2.0;
            include "qelib1.inc";
            qreg q[2];
            creg c[2];
            x q[0];
            cx q[0],q[1];
            measure q[0] -> c[0];
            measure q[1] -> c[1];

        Returns:
            A string representing the OpenQASM2 code corresponding to this
            circuit.
        """
        qiskit_circ = self.subs({}, remove_symbolic=True).to_other_language(
            Language.QISKIT
        )
        assert isinstance(qiskit_circ, QuantumCircuit)
        qasm = qiskit_circ.qasm()
        assert qasm is not None
        return qasm

    def to_qasm3(self) -> str:
        """Converts this circuit to the corresponding OpenQASM 3 code.

        For now, we use an intermediate conversion to OpenQASM 2, and then a
        converter from 2 to 3.

        Example:
            >>> circuit = QCircuit([X(0), CNOT(0, 1), BasisMeasure([0, 1], shots=100)])
            >>> print(circuit.to_qasm3())
            OPENQASM 3.0;
            include 'stdgates.inc';
            qubit[2] q;
            bit[2] c;
            x q[0];
            cx q[0],q[1];
            c[0] = measure q[0];
            c[1] = measure q[1];

        Returns:
            A string representing the OpenQASM3 code corresponding to this
            circuit.
        """
        qasm2_code = self.to_qasm2()
        qasm3_code = open_qasm_2_to_3(qasm2_code)
        return qasm3_code

    def subs(
        self, values: dict[Expr | str, Complex], remove_symbolic: bool = False
    ) -> QCircuit:
        r"""Substitute the parameters of the circuit with complex values.
        Optionally also remove all symbolic variables such as `\pi` (needed for
        example for circuit execution).

        Since we use ``sympy`` for gates' parameters, ``values`` can in fact be
        anything the ``subs`` method from ``sympy`` would accept.

        Example:
            >>> theta, k = symbols("θ k")
            >>> c = QCircuit(
            ...     [Rx(theta, 0), CNOT(1,0), CNOT(1,2), X(2), Rk(2,1), H(0), CRk(k, 0, 1),
            ...      BasisMeasure(list(range(3)), shots=1000)]
            ... )
            >>> print(c)
                 ┌───────┐┌───┐┌───┐                              ┌─┐
            q_0: ┤ Rx(θ) ├┤ X ├┤ H ├────────────■─────────────────┤M├───
                 └───────┘└─┬─┘└───┘┌─────────┐ │P(2**(1 - k)*pi) └╥┘┌─┐
            q_1: ───────────■────■──┤ P(pi/2) ├─■──────────────────╫─┤M├
                               ┌─┴─┐└──┬───┬──┘        ┌─┐         ║ └╥┘
            q_2: ──────────────┤ X ├───┤ X ├───────────┤M├─────────╫──╫─
                               └───┘   └───┘           └╥┘         ║  ║
            c: 3/═══════════════════════════════════════╩══════════╩══╩═
                                                        2          0  1
            >>> print(c.subs({theta: np.py, k: 1}))
                 ┌───────┐┌───┐┌───┐                 ┌─┐
            q_0: ┤ Rx(π) ├┤ X ├┤ H ├───────────■─────┤M├───
                 └───────┘└─┬─┘└───┘┌────────┐ │P(π) └╥┘┌─┐
            q_1: ───────────■────■──┤ P(π/2) ├─■──────╫─┤M├
                               ┌─┴─┐└─┬───┬──┘  ┌─┐   ║ └╥┘
            q_2: ──────────────┤ X ├──┤ X ├─────┤M├───╫──╫─
                               └───┘  └───┘     └╥┘   ║  ║
            c: 3/════════════════════════════════╩════╩══╩═
                                                 2    0  1

        Args:
            values: Mapping between the variables and the replacing values.
            remove_symbolic: If symbolic values should be replaced by their
                numeric counterpart.

        Returns:
            The circuit with the replaced parameters.
        """
        return QCircuit(
            data=[inst.subs(values, remove_symbolic) for inst in self.instructions],
            nb_qubits=self.nb_qubits,
            nb_cbits=self.nb_cbits,
            label=self.label,
        )

    def pretty_print(self):
        """Provides a pretty print of the QCircuit.

        Example:
            >>> c = QCircuit([H(0), CNOT(0,1)])
            >>> c.pretty_print()
            QCircuit : Size (Qubits,Cbits) = (2, None), Nb instructions = 2
                 ┌───┐
            q_0: ┤ H ├──■──
                 └───┘┌─┴─┐
            q_1: ─────┤ X ├
                      └───┘
        """
        print(
            f"QCircuit {self.label or ''}: Size (Qubits,Cbits) = {self.size()},"
            f" Nb instructions = {len(self)}\n"
            f"{self.to_other_language(Language.QISKIT)}"
        )

    def __str__(self) -> str:
        return str(self.to_other_language(Language.QISKIT))

    def __repr__(self) -> str:
        return f"QCircuit({self.instructions})"

    def variables(self):
        """Returns all the parameters involved in this circuit.

        Example:
            >>> circ = QCircuit([
            ...     Rx(theta, 0), CNOT(1,0), CNOT(1,2), X(2), Rk(2,1),
            ...     H(0), CRk(k, 0, 1), ExpectationMeasure([1], obs)
            ... ])
            >>> circ.variables()
            {k, θ}

        Returns:
            All the parameters of the circuit.
        """
        params: set[Basic] = set()
        for inst in self.instructions:
            if isinstance(inst, ParametrizedGate):
                for param in inst.parameters:
                    if isinstance(param, Expr):
                        params.update(param.free_symbols)
        return params



from typing import Optional

import pytest
from qiskit.circuit import Parameter

from mpqp.core.instruction.measurement import Measure
from mpqp.core.languages import Language
from mpqp.tools.generics import flatten


class CustomMeasure(Measure):
    def __init__(self, targets: list[int], shots: int = 0, label: Optional[str] = None):
        if shots < 0:
            raise ValueError(f"Negative number of shots makes no sense, given {shots}")

        super().__init__(targets, shots, label)
        self.c_targets = []

    def to_other_language(
        self,
        language: Language = Language.QISKIT,
        qiskit_parameters: Optional[set[Parameter]] = None,
    ):
        if language == Language.QISKIT:
            pass
        else:
            raise ValueError(f"Unsupported language: {language}")

    @property
    def nb_cbits(self):
        return len(flatten(self.c_targets))


def test_measure_init():
    targets = [0, 1, 2]
    shots = 1
    label = "test_measure"
    measure = CustomMeasure(targets, shots, label)

    assert measure.targets == targets
    assert measure.shots == shots
    assert measure.label == label


def test_negative_shots_raises_error():
    with pytest.raises(ValueError):
        CustomMeasure([0, 1], shots=-1)


def test_nb_qubits():
    measure = CustomMeasure([0, 1])
    assert measure.nb_qubits == 2


def test_nb_cbits_multiple_c_targets():
    custom_measure = CustomMeasure([0, 1], shots=10, label="test_measure")
    custom_measure.c_targets = [[0, 1], [2, 3]]
    assert custom_measure.nb_cbits == len(flatten(custom_measure.c_targets))



class InstructionParsingError(ValueError):
    """Raised when an QASM instruction encountered by the parser is malformed"""


class NumberQubitsError(ValueError):
    """Raised when the number of qubits defining an instruction, a gate, or a measurement, is not coherent with the
    related objets (circuit, matrix, observable, etc.)"""


class ResultAttributeError(AttributeError):
    """Raised when one tries to access the attribute of the result that is incoherent with the associated job."""


class DeviceJobIncompatibleError(ValueError):
    """Raised when one tries to run a job with a JobType that is not suitable for the selected device
    (for example SAMPLE job on a statevector simulator)"""


class RemoteExecutionError(ConnectionError):
    """Raised when an error occurred during a remote connection, submission or execution"""


class IBMRemoteExecutionError(RemoteExecutionError):
    """Raised when an error occurred during the remote execution process of job(s) on an IBM device"""


class QLMRemoteExecutionError(RemoteExecutionError):
    """Raised when an error occurred during the remote execution process of job(s) on the remote QLM"""


class AWSBraketRemoteExecutionError(RemoteExecutionError):
    """Raised when an error occurred during the remote execution process of job(s) on the remote Amazon Braket"""



"""Example 2: Expectation value of an observable"""

import numpy as np
from mpqp.gates import H, Rx
from mpqp import QCircuit
from mpqp.execution import run
from mpqp.measures import ExpectationMeasure, Observable
from mpqp.execution.devices import ATOSDevice, IBMDevice, AWSDevice

obs = Observable(
    np.array(
        [
            [0.63, 0.5, 1, 1],
            [0.5, 0.82, 1, 1],
            [1, 1, 1, 0.33],
            [1, 1, 0.33, 0.3],
        ],
        dtype=float,
    )
)

obs2 = Observable(
    np.array(
        [
            [3.0, 0.0, 0.0, 0.0],
            [0.0, 1.0, 0.0, 0.0],
            [0.0, 0.0, -1.0, 0.0],
            [0.0, 0.0, 0.0, 0.0],
        ],
        dtype=float,
    )
)


# Declaration of the circuit with the right size
circuit = QCircuit(2, label="Observable test")
# Constructing the circuit by adding gates and measurements
circuit.add(H(0))
circuit.add(Rx(1.76, 1))
circuit.add(ExpectationMeasure([0, 1], observable=obs, shots=1000))

print(circuit)

# Running the computation on myQLM and on Aer simulator, then retrieving the results
results = run(circuit, [ATOSDevice.MYQLM_PYLINALG,
                        IBMDevice.AER_SIMULATOR,
                        ATOSDevice.MYQLM_CLINALG,
                        AWSDevice.BRAKET_LOCAL_SIMULATOR])
print(results)


# Declaration of the circuit with the right size
circuit = QCircuit(2, label="Observable test 2")
# Constructing the circuit by adding gates and measurements
circuit.add(H(0))
circuit.add(H(1))
circuit.add(ExpectationMeasure([0, 1], observable=obs2, shots=0))

print(circuit)

# Running the computation on myQLM and on Aer simulator, then retrieving the results
results = run(circuit, [ATOSDevice.MYQLM_PYLINALG,
                        ATOSDevice.MYQLM_CLINALG,
                        IBMDevice.AER_SIMULATOR,
                        AWSDevice.BRAKET_LOCAL_SIMULATOR])
print(results)



"""This module takes care of saving and loading the configuration of supported
providers."""

import os
from dotenv import set_key, load_dotenv
from typeguard import typechecked

MPQP_CONFIG_PATH = os.path.expanduser("~") + "/.mpqp"


def _create_config_if_needed():
    """If there is not already a ``.mpqp`` file we create it."""
    if not os.path.exists(MPQP_CONFIG_PATH):
        open(MPQP_CONFIG_PATH, "a").close()


def get_existing_config_str() -> str:
    """Gets the content of the ``.mpqp`` config file

    Example:
        >>> get_existing_config_str()
        IBM_TOKEN='e7c9*************'
        IBM_CONFIGURED='True'
        QLM_USER='hjaffali'
        QLM_PASSWD='****************'
        QLM_CONFIGURED='True'
        BRAKET_CONFIGURED='True'

    Returns:
        The string with .mpqp file content.
    """
    with open(MPQP_CONFIG_PATH, "r") as mpqp:
        file_str = mpqp.read()
    return file_str


def load_env_variables() -> bool:
    """Loads the variables stored in the ``.mpqp`` file.

    Example:
        >>> os.getenv("IBM_CONFIGURED")

        >>> load_env_variables()
        True
        >>> os.getenv("IBM_CONFIGURED")
        'True'

    Returns:
        ``True`` if the variables are loaded correctly.
    """
    load_dotenv(MPQP_CONFIG_PATH, override=True)
    return True


@typechecked
def get_env_variable(key: str) -> str:
    """Loads the ``.mpqp`` env file and returns the value associated with the key
    in parameter. If the variable does not exist, an empty string is returned.

    Example:
        >>> get_env_variable("BRAKET_CONFIGURED")
        'True'
        >>> get_env_variable("RaNdOM")
        ''

    Args:
        key: The key for which we want to get the value.
    """
    _create_config_if_needed()
    load_env_variables()
    val = os.getenv(key, "")

    return val


@typechecked
def save_env_variable(key: str, value: str) -> bool:
    """Adds or updates the ``key`` environment variable in ``.mpqp`` file.

    Examples:
        >>> get_env_variable("RaNdOM")
        ''
        >>> save_env_variable("RaNdOM", "azertyuiop")
        True
        >>> get_env_variable("RaNdOM")
        'azertyuiop'

    Args:
        key: Name of the environment variable.
        value: Value to be saved.

    Returns:
        ``True`` if the save was successful.
    """
    _create_config_if_needed()

    try:
        a, _, _ = set_key(MPQP_CONFIG_PATH, key, value)
        if a is None:
            raise SystemError(
                "Something went wrong when trying to modify the MPQP "
                "connections configuration."
            )
        load_env_variables()
    except ValueError as e:
        print(e)
        return False

    return a



# pyright: reportUnusedImport=false
from .result import Result, StateVector, Sample, BatchResult
from .runner import run, submit, adjust_measure
from .job import Job, JobStatus, JobType
from .devices import ATOSDevice, AWSDevice, IBMDevice, AvailableDevice
from .remote_handler import remote_result_from_id, remote_result_from_job



from enum import Enum, auto
from abc import abstractmethod

from mpqp.execution.connection.env_manager import get_env_variable


class AvailableDevice(Enum):
    """Class used to define a generic device (quantum computer, or simulator)."""

    @abstractmethod
    def is_remote(self) -> bool:
        """Indicates whether a device is remote or not.

        Returns:
            ``True`` if this device is remote.
        """
        pass

    @abstractmethod
    def is_gate_based(self) -> bool:
        """Indicates whether a device is gate based or not.

        Returns:
            ``True`` if this device is a gate-based simulator/QPU."""
        pass

    @abstractmethod
    def is_simulator(self) -> bool:
        """Indicates whether a device is a simulator or not.

        Returns:
            ``True`` if this device is a simulator."""
        pass


class IBMDevice(AvailableDevice):
    """Enum regrouping all available devices provided by IBM Quantum."""

    PULSE_SIMULATOR = "pulse_simulator"

    AER_SIMULATOR = "aer_simulator"
    AER_SIMULATOR_STATEVECTOR = "aer_simulator_statevector"
    AER_SIMULATOR_DENSITY_MATRIX = "aer_simulator_density_matrix"
    AER_SIMULATOR_STABILIZER = "aer_simulator_stabilizer"
    AER_SIMULATOR_MATRIX_PRODUCT_STATE = "aer_simulator_matrix_product_state"
    AER_SIMULATOR_EXTENDED_STABILIZER = "aer_simulator_extended_stabilizer"
    AER_SIMULATOR_UNITARY = "aer_simulator_unitary"
    AER_SIMULATOR_SUPEROP = "aer_simulator_superop"

    IBMQ_SIMULATOR_STATEVECTOR = "simulator_statevector"
    IBMQ_SIMULATOR_STABILIZER = "simulator_stabilizer"
    IBMQ_SIMULATOR_EXTENDED_STABILIZER = "simulator_extended_stabilizer"
    IBMQ_SIMULATOR_MPS = "simulator_mps"
    IBMQ_QASM_SIMULATOR = "ibmq_qasm_simulator"

    IBM_BRISBANE = "ibm_brisbane"
    IBM_OSAKA = "ibm_osaka"
    IBM_KYOTO = "ibm_kyoto"

    IBM_SHERBROOKE = "ibm_sherbrooke"
    IBM_KYIV = "ibm_kyiv"
    IBM_NAZCA = "ibm_nazca"
    IBM_CUSCO = "ibm_cusco"
    IBM_ITHACA = "ibm_ithaca"
    IBM_TORINO = "ibm_torino"
    IBM_QUEBEC = "ibm_quebec"
    IBM_KAWASAKI = "ibm_kawasaki"
    IBM_CLEVELAND = "ibm_cleveland"
    IBM_CAIRO = "ibm_cairo"
    IBM_HANOI = "ibm_hanoi"
    IBM_ALGIERS = "ibm_algiers"
    IBM_KOLKATA = "ibm_kolkata"
    IBM_MUMBAI = "ibm_mumbai"
    IBM_PEEKSKILL = "ibm_peekskill"

    IBM_RANDOM_SMALL_DEVICE = "ibm_small_device"
    IBM_SMALL_DEVICES_LEAST_BUSY = "ibm_least_busy"

    def is_remote(self) -> bool:
        return self.name.startswith("IBM")

    def is_gate_based(self) -> bool:
        return self != IBMDevice.PULSE_SIMULATOR

    def is_simulator(self) -> bool:
        return "simulator" in self.value


class ATOSDevice(AvailableDevice):
    """Enum regrouping all available devices provided by ATOS."""

    MYQLM_PYLINALG = auto()
    MYQLM_CLINALG = auto()

    QLM_LINALG = auto()
    QLM_MPS = auto()
    QLM_MPS_LEGACY = auto()
    QLM_MPO = auto()
    QLM_STABS = auto()
    QLM_FEYNMAN = auto()
    QLM_BDD = auto()
    QLM_NOISY_QPROC = auto()
    QLM_SQA = auto()
    QLM_QPEG = auto()
    QLM_CLASSICAL_QPU = auto()

    def is_remote(self):
        return self.name.startswith("QLM")

    def is_gate_based(self) -> bool:
        return True

    def is_simulator(self) -> bool:
        return True


class AWSDevice(AvailableDevice):
    """Enum regrouping all available devices provided by AWS Braket."""

    BRAKET_LOCAL_SIMULATOR = "LocalSimulator"

    BRAKET_SV1_SIMULATOR = "quantum-simulator/amazon/sv1"
    BRAKET_DM1_SIMULATOR = "quantum-simulator/amazon/dm1"
    BRAKET_TN1_SIMULATOR = "quantum-simulator/amazon/tn1"

    BRAKET_IONQ_HARMONY = "qpu/ionq/Harmony"
    BRAKET_IONQ_ARIA_1 = "qpu/ionq/Aria-1"
    BRAKET_IONQ_ARIA_2 = "qpu/ionq/Aria-2"
    BRAKET_IONQ_FORTE_1 = "qpu/ionq/Forte-1"
    BRAKET_OQC_LUCY = "qpu/oqc/Lucy"
    BRAKET_QUERA_AQUILA = "qpu/quera/Aquila"
    BRAKET_RIGETTI_ASPEN_M_3 = "qpu/rigetti/Aspen-M-3"

    def is_remote(self):
        return self != AWSDevice.BRAKET_LOCAL_SIMULATOR

    def is_gate_based(self) -> bool:
        return True

    def is_simulator(self) -> bool:
        return "SIMULATOR" in self.name

    def get_arn(self) -> str:
        """
        Retrieve the AwsDevice arn from this AWSDevice element.

        Examples:
            >>> AWSDevice.BRAKET_IONQ_HARMONY.get_arn()
            'arn:aws:braket:us-east-1::device/qpu/ionq/Harmony'
            >>> AWSDevice.BRAKET_SV1_SIMULATOR.get_arn()
            'arn:aws:braket:::device/quantum-simulator/amazon/sv1'
            >>> AWSDevice.BRAKET_RIGETTI_ASPEN_M_3.get_arn()
            'arn:aws:braket:us-west-1::device/qpu/rigetti/Aspen-M-3'

        Returns:
            a string representing the arn of the device
        """
        if not self.is_remote():
            raise ValueError("No arn for a local simulator")
        if self.is_simulator():
            region = ""
        elif self == AWSDevice.BRAKET_RIGETTI_ASPEN_M_3:
            region = "us-west-1"
        elif self == AWSDevice.BRAKET_OQC_LUCY:
            region = "eu-west-2"
        elif self in [
            AWSDevice.BRAKET_IONQ_HARMONY,
            AWSDevice.BRAKET_IONQ_ARIA_1,
            AWSDevice.BRAKET_IONQ_ARIA_2,
            AWSDevice.BRAKET_IONQ_FORTE_1,
            AWSDevice.BRAKET_QUERA_AQUILA,
        ]:
            region = "us-east-1"
        else:
            region = get_env_variable("AWS_DEFAULT_REGION")
        return "arn:aws:braket:" + region + "::device/" + self.value

    @staticmethod
    def from_arn(arn: str):
        """
        Returns the right AWSDevice from the arn given in parameter.

        Examples:
            >>> AWSDevice.from_arn('arn:aws:braket:us-east-1::device/qpu/ionq/Harmony')
            <AWSDevice.BRAKET_IONQ_HARMONY: 'qpu/ionq/Harmony'>
            >>> AWSDevice.from_arn('arn:aws:braket:::device/quantum-simulator/amazon/sv1')
            <AWSDevice.BRAKET_SV1_SIMULATOR: 'quantum-simulator/amazon/sv1'>

        Args:
            arn: The AWS arn identifying the AwsDevice.

        """
        for elem in AWSDevice:
            if elem.value in arn:
                return elem
        return None



from unittest.mock import patch

from mpqp.execution.devices import ATOSDevice, AWSDevice, IBMDevice


def test_ibm_device():
    assert IBMDevice.PULSE_SIMULATOR.is_remote() is False
    assert IBMDevice.PULSE_SIMULATOR.is_gate_based() is False
    assert IBMDevice.PULSE_SIMULATOR.is_simulator() is True

    assert IBMDevice.IBM_BRISBANE.is_remote() is True
    assert IBMDevice.IBM_BRISBANE.is_gate_based() is True
    assert IBMDevice.IBM_BRISBANE.is_simulator() is False


def test_atos_device():
    assert ATOSDevice.QLM_LINALG.is_remote() is True
    assert ATOSDevice.QLM_LINALG.is_gate_based() is True
    assert ATOSDevice.QLM_LINALG.is_simulator() is True

    assert ATOSDevice.MYQLM_PYLINALG.is_gate_based() is True
    assert ATOSDevice.MYQLM_PYLINALG.is_simulator() is True

    assert ATOSDevice.QLM_MPS.is_remote() is True
    assert ATOSDevice.QLM_MPS.is_gate_based() is True
    assert ATOSDevice.QLM_MPS.is_simulator() is True

    assert ATOSDevice.MYQLM_CLINALG.is_gate_based() is True
    assert ATOSDevice.MYQLM_CLINALG.is_simulator() is True


def test_aws_device():
    local_simulator = AWSDevice.BRAKET_LOCAL_SIMULATOR
    assert local_simulator.is_remote() is False
    assert local_simulator.is_gate_based() is True
    assert local_simulator.is_simulator() is True

    harmony_device = AWSDevice.BRAKET_IONQ_HARMONY
    assert harmony_device.is_remote() is True
    assert harmony_device.is_gate_based() is True
    assert harmony_device.is_simulator() is False

    with patch(
        "mpqp.execution.connection.env_manager.get_env_variable",
        return_value="us-east-1",
    ):
        assert (
            harmony_device.get_arn()
            == "arn:aws:braket:us-east-1::device/qpu/ionq/Harmony"
        )
    arn = "arn:aws:braket:::device/quantum-simulator/amazon/sv1"
    assert AWSDevice.from_arn(arn) == AWSDevice.BRAKET_SV1_SIMULATOR


def test_env_manager():
    with patch(
        "mpqp.execution.connection.env_manager.get_env_variable",
        return_value="us-west-1",
    ):
        assert (
            AWSDevice.BRAKET_RIGETTI_ASPEN_M_3.get_arn()
            == "arn:aws:braket:us-west-1::device/qpu/rigetti/Aspen-M-3"
        )



from mpqp import QCircuit
from mpqp.core.instruction.measurement import Observable, ExpectationMeasure
from mpqp.execution.devices import AWSDevice, ATOSDevice
from mpqp.gates import *
from mpqp.measures import BasisMeasure
from mpqp.execution import Sample, run
from mpqp.qasm import qasm3_to_braket_Program, open_qasm_2_to_3
from braket.devices import LocalSimulator
from braket.tasks import GateModelQuantumTaskResult
import numpy as np
from mpqp.qasm.qasm_to_braket import qasm3_to_braket_Circuit

device = LocalSimulator()


qasm_str = """OPENQASM 3.0;
include 'stdgates.inc';
qubit[2] q;
bit c;
h q[0];
cx q[0],q[1];
c[0] = measure q[0];
c[1] = measure q[1];"""


circuit = qasm3_to_braket_Circuit(qasm_str)
print(circuit)
result = device.run(circuit, shots=100).result()
print(f"Measurement counts: {result.measurement_counts}")

#####################################################

# Declaration of the circuit with the right size
circuit = QCircuit(4)

# Constructing the circuit by adding gates
circuit.add(T(0))
circuit.add(CNOT(0, 1))
circuit.add(X(0))
circuit.add(H(1))
circuit.add(Z(2))
circuit.add(CZ(2, 1))
circuit.add(SWAP(2, 0))
circuit.add(CNOT(0, 2))
circuit.add(Ry(3.14 / 2, 2))
circuit.add(S(1))
circuit.add(H(3))
circuit.add(CNOT(1, 2))
circuit.add(Rx(3.14, 1))
circuit.add(CNOT(3, 0))
circuit.add(Rz(3.14, 0))
# Add measurement
circuit.add(BasisMeasure([0, 1, 2, 3], shots=2000))

res = run(circuit, AWSDevice.BRAKET_LOCAL_SIMULATOR)
print(res)

#####################################################

obs = Observable(
    np.array(
        [
            [0.63, 0.5, 1, 1],
            [0.5, 0.82, 1, 1],
            [1, 1, 1, 0.33],
            [1, 1, 0.33, 0.3],
        ],
        dtype=float,
    )
)

# Declaration of the circuit with the right size
circuit = QCircuit(2, label="Observable test")
# Constructing the circuit by adding gates and measurements
circuit.add(H(0))
circuit.add(Rx(1.76, 1))
circuit.add(ExpectationMeasure([0, 1], observable=obs, shots=0))

# Running the computation on myQLM and on Aer simulator, then retrieving the results
result = run(circuit, [AWSDevice.BRAKET_LOCAL_SIMULATOR, ATOSDevice.MYQLM_PYLINALG])
print(result)

#####################################################

# Declaration of the circuit with the right size
circuit = QCircuit(2, label="Observable test")
# Constructing the circuit by adding gates and measurements
circuit.add(H(0))
circuit.add(Rx(1.76, 1))
circuit.add(Ry(1.76, 1))
circuit.add(Rz(1.987, 0))

# Running the computation on myQLM and on Aer simulator, then retrieving the results
result = run(circuit, [AWSDevice.BRAKET_LOCAL_SIMULATOR, ATOSDevice.MYQLM_PYLINALG])
print(result)





"""An :class:`Instruction` is the base element for circuits elements, containing
common methods to all of them."""

from __future__ import annotations

from abc import ABC, abstractmethod
from copy import deepcopy
from typing import Any, Optional
from numbers import Complex

from sympy import Expr
from qiskit.circuit import Parameter
from typeguard import typechecked

from mpqp.core.languages import Language
from mpqp.tools.generics import flatten


@typechecked
class Instruction(ABC):
    """Abstract class defining an instruction of a quantum circuit.

    An Instruction is the elementary component of a
    :class:`QCircuit<mpqp.core.circuit>`. It consists in a manipulation of one
    (or several) qubit(s) of the quantum circuit. It may involve classical bits
    as well, for defining or retrieving the result of the instruction.

    It can be of type:

        - :class:`Gate<mpqp.core.instruction.gates.gate.Gate>`
        - :class:`Measure<mpqp.core.instruction.measurement.measure.Measure>`
        - :class:`Barrier<mpqp.core.instruction.barrier.Barrier>`

    Args:
        targets: List of indices referring to the qubits on which the
            instruction will be applied
        label: label used to identify the instruction
    """

    def __init__(
        self,
        targets: list[int],
        label: Optional[str] = None,
    ):
        if len(set(targets)) != len(targets):
            raise ValueError(f"Duplicate registers in targets: {targets}")
        if not all([t >= 0 for t in targets]):
            raise ValueError(f"Negative index in targets: {targets}")
        self.targets = targets
        """See parameter description."""
        self.label = label
        """See parameter description."""

    @property
    def nb_qubits(self) -> int:
        """Number of qubits of this instruction"""
        return len(self.connections())

    @property
    def nb_cbits(self) -> int:
        """Number of cbits of this instruction"""
        from mpqp.core.instruction.measurement.basis_measure import BasisMeasure

        if isinstance(self, BasisMeasure):
            return len(flatten([self.c_targets]))
        else:
            return 0

    @abstractmethod
    def to_other_language(
        self,
        language: Language = Language.QISKIT,
        qiskit_parameters: Optional[set[Parameter]] = None,
    ) -> Any:
        """Transforms this instruction into the corresponding object in the
        language specified in the ``language`` arg.

        By default, the instruction is translated to the corresponding one in
        Qiskit, since it is the interface we use to generate the OpenQASM code.

        In the future, we will generate the OpenQASM code on our own, and this
        method will be used only for complex objects that are not tractable by
        OpenQASM (like hybrid structures).

        Args:
            language: enum representing the target language.
            qiskit_parameters: We need to keep track of the parameters
                passed to qiskit in order not to define twice the same
                parameter. Defaults to ``set()``.

        Returns:
            The corresponding instruction (gate or measure) in the target
            language
        """
        pass

    def __str__(self) -> str:
        from mpqp.core.circuit import QCircuit

        c = QCircuit(
            (self.targets if isinstance(self.targets, int) else max(self.targets)) + 1
        )
        c.add(self)
        return str(c)

    def __repr__(self) -> str:
        from mpqp.core.instruction.gates import ControlledGate

        controls = str(self.controls) + "," if isinstance(self, ControlledGate) else ""
        return f"{type(self).__name__}({controls}{self.targets})"

    def connections(self) -> set[int]:
        """Returns the indices of the qubits connected to the instruction.

        Example:
            >>> CNOT(0,1).connections()
            [0, 1]

        Returns:
            The qubits ordered connected to instruction.
        """
        from mpqp.core.instruction.gates import ControlledGate

        return (
            set(self.controls).union(self.targets)
            if isinstance(self, ControlledGate)
            else set(self.targets)
        )

    def subs(
        self, values: dict[Expr | str, Complex], remove_symbolic: bool = False
    ) -> Instruction:
        r"""Substitutes the parameters of the instruction with complex values.
        Optionally also removes all symbolic variables such as `\pi` (needed for
        example for circuit execution).

        Since we use ``sympy`` for gates' parameters, ``values`` can in fact be
        anything the ``subs`` method from ``sympy`` would accept.

        Example:
            >>> theta = symbols("θ")
            >>> print(Rx(theta, 0).subs({theta: np.pi}))
               ┌───────┐
            q: ┤ Rx(π) ├
               └───────┘

        Args:
            values: Mapping between the variables and the replacing values.
            remove_symbolic: If symbolic values should be replaced by their
                numeric counterpart.

        Returns:
            The circuit with the replaced parameters.
        """
        return deepcopy(self)



from typing import Optional
from unittest.mock import MagicMock

import numpy as np
import numpy.typing as npt
import pytest
from qiskit.circuit import Parameter

from mpqp.core.instruction.gates.controlled_gate import ControlledGate
from mpqp.core.instruction.gates.gate import Gate
from mpqp.core.languages import Language


class CustomControlledGate(ControlledGate):
    def to_other_language(
        self,
        language: Language = Language.QISKIT,
        qiskit_parameters: Optional[set[Parameter]] = None,
    ):
        assert self.non_controlled_gate is not None
        return self.non_controlled_gate.to_other_language(language, qiskit_parameters)

    def to_matrix(self) -> npt.NDArray[np.complex64]:
        return np.array([[1, 0], [0, 1]], dtype=np.complex64)


@pytest.fixture
def gate_mock() -> MagicMock:
    return MagicMock(spec=Gate)


@pytest.fixture
def controlled_gate(gate_mock: MagicMock):
    controls = [0]
    targets = [1]
    return CustomControlledGate(controls, targets, non_controlled_gate=gate_mock)


def test_init(controlled_gate: ControlledGate, gate_mock: MagicMock):
    assert controlled_gate.controls == [0]
    assert controlled_gate.targets == [1]

    gate_mock.to_matrix.return_value = [[1, 0], [0, 1]]
    gate_mock.to_other_language.return_value = "Mock representation"

    assert np.array_equal(controlled_gate.to_matrix(), np.array([[1, 0], [0, 1]]))
    assert np.array_equal(controlled_gate.to_matrix(), np.array([[1, 0], [0, 1]]))
    assert controlled_gate.to_other_language() == "Mock representation"


@pytest.mark.parametrize(
    "controls, targets, label",
    [([0], [1], "CX"), ([1], [0], "X"), ([0, 1], [2, 3], "H")],
)
def test_controlled_gate_label(
    gate_mock: MagicMock, controls: list[int], targets: list[int], label: str
):
    controlled_gate = CustomControlledGate(
        controls, targets, non_controlled_gate=gate_mock, label=label
    )
    assert controlled_gate.label == label



from __future__ import annotations

# import math
# import numpy as np

# from qat.hardware.default import DefaultGatesSpecification, HardwareModel
# from qiskit_aer.noise import NoiseModel

# from .noise_methods import GateNoise, GateNoiseCombination, NoiseModules
from mpqp.core.languages import Language


class MPQPNoiseModel:
    def __init__(self):
        # 3M-TODO : implement and comment
        self.pp = 1

    def no_idea_what_i_am_doing(
        self, language: Language = Language.QISKIT
    ):  # can be also in MYQLM
        pass



"""File regrouping all features for translating QASM code to Qiskit objects """
from qiskit import QuantumCircuit
from typeguard import typechecked


@typechecked
def qasm2_to_QuantumCircuit(qasm_str: str) -> QuantumCircuit:
    """
    Converting a OpenQASM 2.0 code into a Qiskit QuantumCircuit.

    Args:
        qasm_str: A string representing the OpenQASM 2.0 code.

    Returns:
        A QuantumCircuit equivalent to the QASM code in parameter.
    """
    return QuantumCircuit.from_qasm_str(qasm_str)



"""Native gates is the set of all gates natively supported in OpenQASM. Since we
rely on this standard, all of them are indeed implemented. In addition, this
module contains a few abstract classes used to factorize the behaviors common to
a lot of gates. Feel free to use them for your own custom gates!"""

from __future__ import annotations

from abc import ABC
from numbers import Integral
from typing import Optional

import numpy as np
import numpy.typing as npt

from qiskit.circuit import Parameter
from qiskit.circuit.library import (
    ZGate,
    YGate,
    CCXGate,
    PhaseGate,
    SGate,
    UGate,
    TGate,
    SwapGate,
    RXGate,
    RYGate,
    RZGate,
    HGate,
    CXGate,
    XGate,
    CPhaseGate,
    CZGate,
)
from sympy import Expr, pi

# pylance doesn't handle well Expr, so a lot of "type:ignore" will happen in
# this file :/
from typeguard import typechecked

from mpqp.core.languages import Language
from mpqp.core.instruction.gates.gate import Gate, InvolutionGate, SingleQubitGate
from mpqp.core.instruction.gates.controlled_gate import ControlledGate
from mpqp.core.instruction.gates.gate_definition import UnitaryMatrix
from mpqp.core.instruction.gates.parametrized_gate import ParametrizedGate
from mpqp.tools.generics import Matrix
from mpqp.tools.maths import cos, sin, exp


@typechecked
def _qiskit_parameter_adder(param: Expr | float, qiskit_parameters: set[Parameter]):
    """To avoid having several parameters in qiskit for the same value we keep
    track of them in a set. This function takes care of this, this way you can
    directly call `QiskitGate(_qiskit_parameter_adder(<param>, <q_params_set>))`
    without having to manually take care of the de-duping.

    Args:
        param: The parameter you need for your qiskit gate.
        qiskit_parameters: The set of previously set qiskit parameters. This set
        is updated inplace.

    Returns:

    """
    if isinstance(param, Expr):
        name = str(param)
        previously_set_param = list(
            filter(lambda elt: elt.name == name, qiskit_parameters)
        )
        if len(previously_set_param) > 1:
            raise ReferenceError(
                "Somehow two parameter got the same name, this shouldn't be "
                "possible. For help on this error please contact the authors of"
                " this library"
            )
        elif len(previously_set_param) == 1:
            qiskit_param = previously_set_param[0]
        else:
            qiskit_param = Parameter(name)
            qiskit_parameters.add(qiskit_param)
    else:
        qiskit_param = param
    return qiskit_param


@typechecked
class NativeGate(Gate, ABC):
    """The standard on which we rely, OpenQASM, comes with a set of gates
    supported by default. More complicated gates can be defined by the user.
    This class represent all those gates supported by default.

    Args:
        targets: List of indices referring to the qubits on which the gate will
            be applied.
        label: Label used to identify the gate.
    """

    native_gate_options = {"disable_symbol_warn": True}


@typechecked
class RotationGate(NativeGate, ParametrizedGate, ABC):
    """Many gates can be classified as a simple rotation gate, around a specific
    axis (and potentially with a control qubit). All those gates have in common
    a single parameter: ``theta``. This class help up factorize this behavior,
    and simply having to tweak the matricial semantics and qasm translation of
    the specific gate.

    Args:
        theta: Angle of the rotation.
        target: Index referring to the qubits on which the gate will be applied.
    """

    qiskit_gate: type[RXGate | RYGate | RZGate | PhaseGate | CPhaseGate]

    def __init__(self, theta: Expr | float, target: int):
        self.parameters = [theta]
        definition = UnitaryMatrix(self.to_matrix(), **self.native_gate_options)
        ParametrizedGate.__init__(
            self, definition, [target], [self.theta], type(self).__name__.capitalize()
        )

    @property
    def theta(self):
        return self.parameters[0]

    def __repr__(self):
        return (
            f"{type(self).__name__}({f'{self.theta}'.rstrip('0')}, {self.targets[0]})"
        )

    def to_other_language(
        self,
        language: Language = Language.QISKIT,
        qiskit_parameters: Optional[set[Parameter]] = None,
    ):
        if qiskit_parameters is None:
            qiskit_parameters = set()
        theta = float(self.theta) if self._numeric_parameters else self.theta
        if language == Language.QISKIT:
            return self.qiskit_gate(_qiskit_parameter_adder(theta, qiskit_parameters))
        else:
            raise NotImplementedError(f"Error: {language} is not supported")


@typechecked
class NoParameterGate(NativeGate, ABC):
    """Class describing native gates that do not depend on parameters.

    Args:
        targets: List of indices referring to the qubits on which the gate will
            be applied.
        label: Label used to identify the gate.
    """

    qiskit_gate: type[
        XGate
        | YGate
        | ZGate
        | HGate
        | TGate
        | SGate
        | SwapGate
        | CXGate
        | CZGate
        | CCXGate
    ]
    """Corresponding ``qiskit``'s gate class."""
    matrix: npt.NDArray[np.complex64]
    """Matricial semantics of the gate."""

    def to_other_language(
        self,
        language: Language = Language.QISKIT,
        qiskit_parameters: Optional[set[Parameter]] = None,
    ):
        if language == Language.QISKIT:
            return self.qiskit_gate()
        else:
            raise NotImplementedError(f"Error: {language} is not supported")

    def to_matrix(self) -> Matrix:
        return self.matrix


@typechecked
class OneQubitNoParamGate(SingleQubitGate, NoParameterGate, ABC):
    """Class describing one-qubit native gates that do not depend on parameters.

    Args:
        target: Index referring to the qubits on which the gate will be applied.
    """

    def __init__(self, target: int):
        SingleQubitGate.__init__(self, target, type(self).__name__)


class X(OneQubitNoParamGate, InvolutionGate):
    """One qubit X (NOT) Pauli gate.

    Args:
        target: Index referring to the qubit on which the gate will be applied.

    Example:
        >>> X(0).to_matrix()
        array([[0, 1],
               [1, 0]])
    """

    qiskit_gate = XGate
    matrix = np.array([[0, 1], [1, 0]])


class Y(OneQubitNoParamGate, InvolutionGate):
    """One qubit Y Pauli gate.

    Args:
        target: Index referring to the qubit on which the gate will be applied.

    Example:
        >>> Y(0).to_matrix()
        array([[ 0.+0.j, -0.-1.j],
               [ 0.+1.j,  0.+0.j]])
    """

    qiskit_gate = YGate
    matrix = np.array([[0, -1j], [1j, 0]])


class Z(OneQubitNoParamGate, InvolutionGate):
    """One qubit Z Pauli gate.

    Args:
        target: Index referring to the qubit on which the gate will be applied.

    Example:
        >>> Z(0).to_matrix()
        array([[ 1,  0],
               [ 0, -1]])
    """

    qiskit_gate = ZGate
    matrix = np.array([[1, 0], [0, -1]])


class H(OneQubitNoParamGate, InvolutionGate):
    """One qubit Hadamard gate.

    Args:
        target: Index referring to the qubit on which the gate will be applied.

    Example:
        >>> H(0).to_matrix()
        array([[ 0.70710678,  0.70710678],
               [ 0.70710678, -0.70710678]])
    """

    qiskit_gate = HGate
    matrix = np.array([[1, 1], [1, -1]]) / np.sqrt(2)


class P(RotationGate, SingleQubitGate):
    """One qubit parametrized Phase gate. Consist in a rotation around Z axis.

    Args:
        theta: Parameter representing the phase to apply.
        target: Index referring to the qubit on which the gate will be applied.

    Example:
        >>> P(np.pi/3, 1).to_matrix()
        array([[1. +0.j       , 0. +0.j       ],
               [0. +0.j       , 0.5+0.8660254j]])
    """

    qiskit_gate = PhaseGate

    def to_matrix(self) -> Matrix:
        return np.array([[1, 0], [0, exp(self.parameters[0] * 1j)]])  # type:ignore


class S(OneQubitNoParamGate):
    """One qubit S gate. It's equivalent to ``P(pi/2)``.
    It can also be defined as the square-root of the Z (Pauli) gate.

    Args:
        target: Index referring to the qubit on which the gate will be applied.

    Example:
        >>> S(0).to_matrix()
        array([[1.+0.j, 0.+0.j],
               [0.+0.j, 0.+1.j]])
    """

    qiskit_gate = SGate
    matrix = np.array([[1, 0], [0, 1j]])


class T(OneQubitNoParamGate):
    """One qubit T gate. It consists in applying a phase of Pi/4.
    The T gate can be defined as the fourth-root of the Z (Pauli) gate.

    Args:
        target: Index referring to the qubit on which the gate will be applied.

    Example:
        >>> T(0).to_matrix()
        array([[1.        +0.j        , 0.        +0.j        ],
               [0.        +0.j        , 0.70710678+0.70710678j]])
    """

    qiskit_gate = TGate
    matrix = np.array([[1, 0], [0, exp((pi / 4) * 1j)]])


class SWAP(InvolutionGate, NoParameterGate):
    """Two-qubit SWAP gate.

    Args:
        a: First target of the swapping operation.
        b: Second target of the swapping operation.

    Example:
        >>> SWAP(0, 1)
        array([[1, 0, 0, 0],
               [0, 0, 1, 0],
               [0, 1, 0, 0],
               [0, 0, 0, 1]])
    """

    qiskit_gate = SwapGate
    matrix = np.array([[1, 0, 0, 0], [0, 0, 1, 0], [0, 1, 0, 0], [0, 0, 0, 1]])

    def __init__(self, a: int, b: int):
        super().__init__([a, b], "SWAP")


class U(NativeGate, ParametrizedGate):
    """Generic one qubit unitary gate. It is parametrized by 3 Euler angles.

    Args:
        theta: Parameter representing the first angle of the gate U.
        phi: Parameter representing the second angle of the gate U.
        gamma: Parameter representing the third angle of the gate U.
        target: Index referring to the qubit on which the gate will be applied.

    Example:
        >>> U(np.pi/3, 0, np.pi/4, 0).to_matrix()
        array([[ 0.8660254 +0.j        , -0.35355339-0.35355339j],
               [ 0.5       +0.j        ,  0.61237244+0.61237244j]])
    """

    def __init__(
        self,
        theta: Expr | float,
        phi: Expr | float,
        gamma: Expr | float,
        target: int,
    ):
        self.parameters = [theta, phi, gamma]
        definition = UnitaryMatrix(self.to_matrix(), **self.native_gate_options)
        ParametrizedGate.__init__(self, definition, [target], [theta, phi, gamma], "U")

    @property
    def theta(self):
        """See corresponding argument."""
        return self.parameters[0]

    @property
    def phi(self):
        """See corresponding argument."""
        return self.parameters[1]

    @property
    def gamma(self):
        """See corresponding argument."""
        return self.parameters[2]

    def to_other_language(
        self,
        language: Language = Language.QISKIT,
        qiskit_parameters: Optional[set[Parameter]] = None,
    ):
        if qiskit_parameters is None:
            qiskit_parameters = set()
        if language == Language.QISKIT:
            return UGate(
                theta=_qiskit_parameter_adder(self.theta, qiskit_parameters),
                phi=_qiskit_parameter_adder(self.phi, qiskit_parameters),
                lam=_qiskit_parameter_adder(self.gamma, qiskit_parameters),
            )
        else:
            raise NotImplementedError(f"Error: {language} is not supported")

    def to_matrix(self) -> Matrix:
        c, s, eg, ep = (
            cos(self.theta / 2),  # type:ignore
            sin(self.theta / 2),  # type:ignore
            exp(self.gamma * 1j),  # type:ignore
            exp(self.phi * 1j),  # type:ignore
        )
        return np.array([[c, -eg * s], [ep * s, eg * ep * c]])  # type:ignore


class Rx(RotationGate, SingleQubitGate):
    """One qubit rotation around the X axis

    Args:
        theta: Parameter representing the angle of the gate.
        target: Index referring to the qubit on which the gate will be applied.

    Example:
        >>> Rx(np.pi/5, 1).to_matrix()
        array([[0.95105652+0.j        , 0.        -0.30901699j],
               [0.        -0.30901699j, 0.95105652+0.j        ]])
    """

    qiskit_gate = RXGate

    def to_matrix(self) -> Matrix:
        c, s = cos(self.parameters[0] / 2), sin(self.parameters[0] / 2)  # type:ignore
        return np.array([[c, -1j * s], [-1j * s, c]])  # type:ignore


class Ry(RotationGate, SingleQubitGate):
    """One qubit rotation around the Y axis

    Args:
        theta: Parameter representing the angle of the gate.
        target: Index referring to the qubit on which the gate will be applied.

    Example:
        >>> Ry(np.pi/5, 1).to_matrix()
        array([[ 0.95105652, -0.30901699],
               [ 0.30901699,  0.95105652]])
    """

    qiskit_gate = RYGate

    def to_matrix(self) -> Matrix:
        c, s = cos(self.parameters[0] / 2), sin(self.parameters[0] / 2)  # type:ignore
        return np.array([[c, -s], [s, c]])


class Rz(RotationGate, SingleQubitGate):
    """One qubit rotation around the Z axis

    Args:
        theta: Parameter representing the angle of the gate.
        target: Index referring to the qubit on which the gate will be applied.

    Example:
        >>> Rz(np.pi/5, 1).to_matrix()
        array([[0.95105652-0.30901699j, 0.        +0.j        ],
               [0.        +0.j        , 0.95105652-0.30901699j]])
    """

    qiskit_gate = RZGate

    def to_matrix(self) -> Matrix:
        e = exp(-1j * self.parameters[0] / 2)  # type:ignore
        return np.array([[e, 0], [0, e]])


class Rk(RotationGate, SingleQubitGate):
    r"""One qubit Phase gate of angle `\frac{2i\pi}{2^k}`.

    Args:
        k: Parameter used in the definition of the phase to apply.
        target: Index referring to the qubit on which the gate will be applied.

    Example:
        >>> Rk(5, 0).to_matrix()
        array([[1.        +0.j        , 0.        +0.j        ],
               [0.        +0.j        , 0.98078528+0.19509032j]])
    """

    qiskit_gate = PhaseGate

    def __init__(self, k: Expr | int, target: int):
        self.parameters = [k]
        definition = UnitaryMatrix(self.to_matrix(), **self.native_gate_options)
        ParametrizedGate.__init__(self, definition, [target], [self.k], "Rk")

    @property
    def theta(self) -> Expr | float:
        r"""Value of the rotation angle, parametrized by ``k`` with the relation
        `\theta = \frac{\pi}{2^{k-1}}`."""
        return pi / 2 ** (self.k - 1)  # type:ignore

    @property
    def k(self) -> Expr | float:
        """See corresponding argument."""
        return self.parameters[0]

    def to_matrix(self) -> Matrix:
        p = np.pi if isinstance(self.k, Integral) else pi
        e = exp(p * 1j / 2 ** (self.k - 1))  # type:ignore
        return np.array([[1, 0], [0, e]])

    def __repr__(self):
        return f"{type(self).__name__}({self.k}, {self.targets[0]})"


class CNOT(InvolutionGate, NoParameterGate, ControlledGate):
    """
    Two-qubit Controlled-NOT gate.

    Args:
        control: index referring to the qubit used to control the gate
        target: index referring to the qubit on which the gate will be applied

    Example:
        >>> CNOT(0, 1).to_matrix()
        array([[1, 0, 0, 0],
               [0, 1, 0, 0],
               [0, 0, 0, 1],
               [0, 0, 1, 0]])
    """

    qiskit_gate = CXGate

    def __init__(self, control: int, target: int):
        ControlledGate.__init__(self, [control], [target], X(target), "CNOT")

    def to_matrix(self) -> Matrix:
        return np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, 1], [0, 0, 1, 0]])


class CZ(InvolutionGate, NoParameterGate, ControlledGate):
    """Two-qubit Controlled-Z gate.

    Args:
        control: Index referring to the qubit used to control the gate.
        target: Index referring to the qubit on which the gate will be applied.

    Example:
        >>> CZ(0, 1).to_matrix()
        array([[ 1,  0,  0,  0],
               [ 0,  1,  0,  0],
               [ 0,  0,  1,  0],
               [ 0,  0,  0, -1]])
    """

    qiskit_gate = CZGate

    def __init__(self, control: int, target: int):
        ControlledGate.__init__(self, [control], [target], Z(target), "CZ")

    def to_matrix(self) -> Matrix:
        m = np.eye(8, dtype=complex)
        m[-1, -1] = -1
        return m


class CRk(RotationGate, ControlledGate):
    """Two-qubit Controlled-Rk gate.

    Args:
        k: Parameter used in the definition of the phase to apply.
        control: Index referring to the qubit used to control the gate.
        target: Index referring to the qubit on which the gate will be applied.

    Example:
        >>> CRk(4, 0, 1)
        array([[1.        +0.j        , 0.        +0.j        , 0.        +0.j        , 0.        +0.j        ],
               [0.        +0.j        , 1.        +0.j        , 0.        +0.j        , 0.        +0.j        ],
               [0.        +0.j        , 0.        +0.j        , 1.        +0.j        , 1.        +0.j        ],
               [0.        +0.j        , 0.        +0.j        , 0.        +0.j        , 0.92387953+0.38268343j]])
    """

    qiskit_gate = CPhaseGate

    def __init__(self, k: Expr | int, control: int, target: int):
        self.parameters = [k]
        ControlledGate.__init__(self, [control], [target], Rk(k, target), "CRk")
        definition = UnitaryMatrix(self.to_matrix(), **self.native_gate_options)
        ParametrizedGate.__init__(self, definition, [target], [k], "CRk")

    @property
    def theta(self) -> Expr | float:
        r"""Value of the rotation angle, parametrized by ``k`` with the relation
        `\theta = \frac{\pi}{2^{k-1}}`."""
        p = np.pi if isinstance(self.k, Integral) else pi
        return p / 2 ** (self.k - 1)  # type:ignore

    @property
    def k(self) -> Expr | float:
        """See corresponding argument."""
        return self.parameters[0]

    def to_matrix(self) -> Matrix:
        e = exp(self.theta * 1j)  # type:ignore
        return np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, e]])

    def __repr__(self):
        return f"{type(self).__name__}({self.k}, {self.controls[0]}, {self.targets[0]})"


class TOF(InvolutionGate, NoParameterGate, ControlledGate):
    """Three-qubit Controlled-Controlled-NOT gate, also known as Toffoli Gate

    Args:
        control: List of indices referring to the qubits used to control the gate.
        target: Index referring to the qubit on which the gate will be applied.

    Example:
        >>> TOF([0, 1], 2).to_matrix()
        array([[1, 0, 0, 0, 0, 0, 0, 0],
               [0, 1, 0, 0, 0, 0, 0, 0],
               [0, 0, 1, 0, 0, 0, 0, 0],
               [0, 0, 0, 1, 0, 0, 0, 0],
               [0, 0, 0, 0, 1, 0, 0, 0],
               [0, 0, 0, 0, 0, 1, 0, 0],
               [0, 0, 0, 0, 0, 0, 0, 1],
               [0, 0, 0, 0, 0, 0, 1, 0]])
    """

    qiskit_gate = CCXGate

    def __init__(self, control: list[int], target: int):
        if len(control) != 2:
            raise ValueError("A Toffoli gate must have exactly 2 control qubits.")
        ControlledGate.__init__(self, control, [target], X(target), "TOF")

    def to_matrix(self) -> Matrix:
        m = np.identity(8, dtype=complex)
        m[-2:, -2:] = np.ones(2) - np.identity(2)
        return m



"""A barrier is a purely cosmetic instruction. In fact, at execution it is
removed because it could have a negative impact on the execution speed of the
circuit (since it artificially increases the depth)."""

from typing import Optional
from qiskit.circuit.library import Barrier as QiskitBarrier

from qiskit.circuit import Parameter

from mpqp.core.languages import Language
from .instruction import Instruction


class Barrier(Instruction):
    """Visual indicator of the grouping of circuit sections"""

    def __init__(self):
        super().__init__([0])
        self.size = 0
        """Size of the barrier (set to 0 by default)."""

    def to_other_language(
        self,
        language: Language = Language.QISKIT,
        qiskit_parameters: Optional[set[Parameter]] = None,
    ):
        if language == Language.QISKIT:
            return QiskitBarrier(self.size)



from .qasm_to_braket import qasm3_to_braket_Program
from .qasm_to_myqlm import qasm2_to_myqlm_Circuit
from .qasm_to_qiskit import qasm2_to_QuantumCircuit
from .open_qasm_2_and_3 import (
    open_qasm_2_to_3,
    open_qasm_file_conversion_2_to_3,
    open_qasm_hard_includes,
)



# pyright: reportUnusedImport=false
from .atos_execution import run_atos, run_myQLM, run_QLM
from .aws_execution import run_braket, submit_job_braket
from .ibm_execution import run_ibm, run_aer, run_ibmq



# pyright: reportUnusedImport=false
from .controlled_gate import ControlledGate
from .gate import Gate
from .parametrized_gate import ParametrizedGate, symbols
from .gate_definition import (
    GateDefinition,
    KrausRepresentation,
    PauliDecomposition,
)
from .native_gates import (
    X,
    Y,
    Z,
    H,
    P,
    S,
    T,
    SWAP,
    U,
    Rx,
    Ry,
    Rz,
    Rk,
    CNOT,
    CZ,
    CRk,
    TOF,
)
from .custom_gate import CustomGate, UnitaryMatrix



"""File regrouping all features for translating QASM code to Amazon Braket objects."""

from braket.ir.openqasm import Program
from braket.circuits import Circuit
from typeguard import typechecked

from mpqp.qasm.open_qasm_2_and_3 import open_qasm_hard_includes


@typechecked
def qasm3_to_braket_Program(qasm3_str: str) -> Program:
    """Converting a OpenQASM 3.0 code into a Braket Program.

    Args:
        qasm3_str: A string representing the OpenQASM 3.0 code.

    Returns:
        A Program equivalent to the QASM code in parameter.
    """

    # PROBLEM: import and standard gates are not supported by Braket
    # NOTE: however custom OpenQASM 3 gates declaration is supported by Braket,
    # the idea is then to hard import the standard lib and other files into the qasm string's header before
    # giving it to the Program.
    after_stdgates_included = open_qasm_hard_includes(qasm3_str, set())

    program = Program(source=after_stdgates_included, inputs=None)
    return program


@typechecked
def qasm3_to_braket_Circuit(qasm3_str: str) -> Circuit:
    """
    Converting a OpenQASM 3.0 code into a Braket Circuit

    Args:
        qasm3_str: A string representing the OpenQASM 3.0 code.

    Returns:
        A Circuit equivalent to the QASM code in parameter.
    """

    # PROBLEM: import and standard gates are not supported by Braket
    # NOTE: however custom OpenQASM 3 gates declaration is supported by Braket,
    # SOLUTION: the idea is then to hard import the standard lib and other files into the qasm string's header before
    # giving it to the Circuit.
    # PROBLEM2: Braket doesn't support NATIVE freaking gates U and gphase, so the trick may not work for the moment
    # for circuit, only for program
    # SOLUTION: import a specific qasm file with U and gphase redefined with the supported Braket SDK gates, and by
    # removing from this import file the already handled gates

    # we remove any include of stdgates.inc and replace it with custom include
    qasm3_str = qasm3_str.replace("stdgates.inc", "braket_custom_include.inc")

    after_stdgates_included = open_qasm_hard_includes(qasm3_str, set())
    # NOTE : gphase is a already used in Braket and thus cannot be redefined as a native gate in OpenQASM.
    # We used ggphase instead

    circuit = Circuit.from_ir(after_stdgates_included)
    return circuit



import pytest
from mpqp.qasm import open_qasm_file_conversion_2_to_3, open_qasm_hard_includes

qasm_folder = "tests/qasm/qasm_examples/"


def test_qasm_hard_import():
    with_include_fn = qasm_folder + "with_include.qasm"
    with open(with_include_fn) as f1:
        with open(qasm_folder + "without_include.qasm") as f2:
            assert open_qasm_hard_includes(f1.read(), {with_include_fn}) == f2.read()


def test_late_gate_def():
    with pytest.raises(ValueError):
        open_qasm_file_conversion_2_to_3(qasm_folder + "late_gate_def.qasm")


def test_in_time_gate_def():
    file_name = qasm_folder + "in_time_gate_def.qasm"
    converted_file_name = qasm_folder + "in_time_gate_def_converted.qasm"
    with open(converted_file_name, "r") as f:
        assert open_qasm_file_conversion_2_to_3(file_name) == f.read()


def test_circular_dependency_detection():
    with pytest.raises(RuntimeError) as e:
        open_qasm_file_conversion_2_to_3(
            qasm_folder + "circular_dep1.qasm",
        )
    assert "Circular dependency" in str(e.value)


def test_circular_dependency_detection_2():
    try:
        open_qasm_file_conversion_2_to_3(
            qasm_folder + "circular_dep_a.qasm",
        )
    except RuntimeError:
        assert False, f"Circular dependency raised while it shouldn't"



from __future__ import annotations

from numbers import Complex, Real

import numpy as np
import numpy.typing as npt
import sympy as sp
from sympy import Expr, I, pi
from typeguard import typechecked

from mpqp.tools.generics import Matrix

rtol = 1e-05
"""The relative tolerance parameter."""
atol = 1e-08
"""The absolute tolerance parameter."""


@typechecked
def normalize(v: npt.NDArray[np.complex64]) -> npt.NDArray[np.complex64]:
    """Normalizes an array representing the amplitudes of the state.

    Examples:
        >>> vector = np.array([1,0,0,1])
        >>> normalize(vector)
        array([0.70710678, 0., 0., 0.70710678])
        >>> vector = np.array([0,0,0,0])
        >>> normalize(vector)
        array([0, 0, 0, 0])

    Args:
        v: vector to be normalized

    Returns:
        The normalized vector.
    """
    norm = np.linalg.norm(v, ord=2)
    return v if norm == 0 else v / norm


@typechecked
def matrix_eq(lhs: Matrix, rhs: Matrix) -> bool:
    r"""Checks whether two matrix are element-wise equal, within a tolerance.

    For respectively each elements `a` and `b` of both inputs, we check this
    specific condition: `|a - b| \leq (atol + rtol * |b|)`.

    Args:
        lhs: Left-hand side matrix of the equality.
        rhs: Right-hand side matrix of the equality.

    Returns:
        ``True`` if the two matrix are equal (according to the definition above).
    """
    for elt in zip(np.ndarray.flatten(lhs), np.ndarray.flatten(rhs)):
        if isinstance(elt[0], Expr) or isinstance(elt[1], Expr):
            if elt[0] != elt[1]:
                return False
        else:
            if abs(elt[0] - elt[1]) > (atol + rtol * abs(elt[1])):
                return False
    return True


@typechecked
def is_hermitian(matrix: Matrix) -> bool:
    """Checks whether the matrix in parameter is hermitian.

    Args:
        matrix: matrix for which we want to know if it is hermitian

    Examples:
        >>> m1 = np.array([[1,2j,3j],[-2j,4,5j],[-3j,-5j,6]])
        >>> is_hermitian(m1)
        True
        >>> m2 = np.diag([1,2,3,4])
        >>> is_hermitian(m2)
        True
        >>> m3 = np.array([[1,2,3],[2,4,5],[3,5,6]])
        >>> is_hermitian(m3)
        True
        >>> m4 = np.array([[1,2,3],[4,5,6],[7,8,9]])
        >>> is_hermitian(m4)
        False
        >>> x = symbols("x", real=True)
        >>> m5 = np.diag([1,x])
        >>> is_hermitian(m5)
        True
        >>> m6 = np.array([[1,x],[-x,2]])
        >>> is_hermitian(m6)
        False

    Returns:
        ``True`` if the matrix in parameter is Hermitian.
    """
    return matrix_eq(np.array(matrix).transpose().conjugate(), matrix)  # type: ignore


@typechecked
def is_unitary(matrix: Matrix) -> bool:
    """Checks whether the matrix in parameter is unitary.

    Example:
        >>> a = np.array([[1,1],[1,-1]])
        >>> is_unitary(a)
        False
        >>> is_unitary(a/np.sqrt(2))
        True

    Args:
        matrix: Matrix for which we want to know if it is unitary.

    Returns:
        ``True`` if the matrix in parameter is Unitary.
    """
    return matrix_eq(np.eye(len(matrix)), matrix.transpose().conjugate().dot(matrix))


@typechecked
def cos(angle: sp.Expr | float) -> sp.Expr | float:
    """Generalization of the cosine function, to take as input either
    ``sympy``'s expressions or floating numbers.

    Args:
        angle: The angle considered.

    Returns:
        Cosine of the given ``angle``.
    """
    return np.cos(angle) if isinstance(angle, Real) else sp.cos(angle)


@typechecked
def sin(angle: sp.Expr | float) -> sp.Expr | float:
    """Generalization of the sine function, to take as input either
    ``sympy``'s expressions or floating numbers.

    Args:
        angle: The angle considered.

    Returns:
        Sine of the given ``angle``.
    """
    return np.sin(angle) if isinstance(angle, Real) else sp.sin(angle)


@typechecked
def exp(angle: sp.Expr | complex) -> sp.Expr | complex:
    """Generalization of the exponential function, to take as input either
    ``sympy``'s expressions or floating numbers.

    Args:
        angle: The angle considered.

    Returns:
        Exponential of the given ``angle``.
    """
    return np.exp(angle) if isinstance(angle, Complex) else sp.exp(angle)
